#
# Class for particles using the MSMR model
#
import pybamm
from .base_particle import BaseParticle


class MSMRDiffusion(BaseParticle):
    """
    Class for molar conservation in particles within the Multi-Species Multi-Reaction
    framework :footcite:t:`Baker2018`. The thermodynamic model is presented in
    :footcite:t:`Verbrugge2017`, along with parameter values for a number of
    substitutional materials.

    In this submodel, the stoichiometry depends on the potential in the particle and
    the temperature, so dUdT is not used. See `:meth:`pybamm.LithiumIonParameters.dUdT`
    for more explanation.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain of the model either 'Negative' or 'Positive'
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    x_average : bool
        Whether the particle concentration is averaged over the x-direction
    """

    def __init__(self, param, domain, options, phase="primary", x_average=False):
        super().__init__(param, domain, options, phase)
        self.x_average = x_average

        pybamm.citations.register("Baker2018")
        pybamm.citations.register("Verbrugge2017")

    def get_fundamental_variables(self):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        variables = {}

        # Define "particle" potential variables. In the MSMR model, we solve for the
        # potential as a function of position within the electrode and particles (and
        # particle-size distribution, if applicable). The potential is then used to
        # calculate the stoichiometry, which is used to calculate the particle
        # concentration.
        if self.size_distribution is False:
            if self.x_average is False:
                U = pybamm.Variable(
                    f"{Domain} {phase_name}particle potential [V]",
                    f"{domain} {phase_name}particle",
                    auxiliary_domains={
                        "secondary": f"{domain} electrode",
                        "tertiary": "current collector",
                    },
                )
                U.print_name = f"U_{domain[0]}"
            else:
                U_xav = pybamm.Variable(
                    f"X-averaged {domain} {phase_name}particle potential [V]",
                    f"{domain} {phase_name}particle",
                    auxiliary_domains={"secondary": "current collector"},
                )
                U_xav.print_name = f"U_{domain[0]}_xav"
                U = pybamm.SecondaryBroadcast(U_xav, f"{domain} electrode")
        else:
            if self.x_average is False:
                U_distribution = pybamm.Variable(
                    f"{Domain} {phase_name}particle potential distribution [V]",
                    domain=f"{domain} {phase_name}particle",
                    auxiliary_domains={
                        "secondary": f"{domain} {phase_name}particle size",
                        "tertiary": f"{domain} electrode",
                        "quaternary": "current collector",
                    },
                )
                R = pybamm.SpatialVariable(
                    f"R_{domain[0]}",
                    domain=[f"{domain} {phase_name}particle size"],
                    auxiliary_domains={
                        "secondary": f"{domain} electrode",
                        "tertiary": "current collector",
                    },
                    coord_sys="cartesian",
                )
                variables = self._get_distribution_variables(R)
                f_v_dist = variables[
                    f"{Domain} volume-weighted {phase_name}"
                    "particle-size distribution [m-1]"
                ]
            else:
                U_distribution = pybamm.Variable(
                    f"X-averaged {domain} {phase_name}particle "
                    "potential distribution [V]",
                    domain=f"{domain} {phase_name}particle",
                    auxiliary_domains={
                        "secondary": f"{domain} {phase_name}particle size",
                        "tertiary": "current collector",
                    },
                )
                R = pybamm.SpatialVariable(
                    f"R_{domain[0]}",
                    domain=[f"{domain} {phase_name}particle size"],
                    auxiliary_domains={"secondary": "current collector"},
                    coord_sys="cartesian",
                )
                variables = self._get_distribution_variables(R)
                f_v_dist = variables[
                    f"X-averaged {domain} volume-weighted {phase_name}"
                    "particle-size distribution [m-1]"
                ]

            # Standard potential distribution_variables
            variables.update(
                self._get_standard_potential_distribution_variables(U_distribution)
            )

            # Standard size-averaged variables. Average potentials using
            # the volume-weighted distribution since they are volume-based
            # quantities. Necessary for output variables "Total lithium in
            # negative electrode [mol]", etc, to be calculated correctly
            U = pybamm.Integral(f_v_dist * U_distribution, R)
            if self.x_average is True:
                U = pybamm.SecondaryBroadcast(U, [f"{domain} electrode"])

        # Standard potential variables
        variables.update(self._get_standard_potential_variables(U))

        return variables

    def get_coupled_variables(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        if self.size_distribution is False:
            if self.x_average is False:
                x = variables[f"{Domain} {phase_name}particle stoichiometry"]
                dxdU = variables[
                    f"{Domain} {phase_name}particle differential stoichiometry [V-1]"
                ]
                U = variables[f"{Domain} {phase_name}particle potential [V]"]
                T = pybamm.PrimaryBroadcast(
                    variables[f"{Domain} electrode temperature [K]"],
                    [f"{domain} {phase_name}particle"],
                )
                R_nondim = variables[f"{Domain} {phase_name}particle radius"]
                j = variables[
                    f"{Domain} electrode {phase_name}"
                    "interfacial current density [A.m-2]"
                ]
            else:
                x = variables[f"X-averaged {domain} {phase_name}particle stoichiometry"]
                dxdU = variables[
                    f"X-averaged {domain} {phase_name}particle differential "
                    "stoichiometry [V-1]"
                ]
                U = variables[f"X-averaged {domain} {phase_name}particle potential [V]"]
                T = pybamm.PrimaryBroadcast(
                    variables[f"X-averaged {domain} electrode temperature [K]"],
                    [f"{domain} {phase_name}particle"],
                )
                R_nondim = 1
                j = variables[
                    f"X-averaged {domain} electrode {phase_name}"
                    "interfacial current density [A.m-2]"
                ]
            R_broad_nondim = R_nondim
        else:
            R_nondim = variables[f"{Domain} {phase_name}particle sizes"]
            R_broad_nondim = pybamm.PrimaryBroadcast(
                R_nondim, [f"{domain} {phase_name}particle"]
            )
            if self.x_average is False:
                x = variables[
                    f"{Domain} {phase_name}particle stoichiometry distribution"
                ]
                dxdU = variables[
                    f"{Domain} {phase_name}particle differential stoichiometry "
                    "distribution [V-1]"
                ]
                U = variables[
                    f"{Domain} {phase_name}particle potential distribution [V]"
                ]
                # broadcast T to "particle size" domain then again into "particle"
                T = pybamm.PrimaryBroadcast(
                    variables[f"{Domain} electrode temperature [K]"],
                    [f"{domain} {phase_name}particle size"],
                )
                T = pybamm.PrimaryBroadcast(T, [f"{domain} {phase_name}particle"])
                j = variables[
                    f"{Domain} electrode {phase_name}interfacial "
                    "current density distribution [A.m-2]"
                ]
            else:
                x = variables[
                    f"X-averaged {domain} {phase_name}particle "
                    "stoichiometry distribution"
                ]
                dxdU = variables[
                    f"X-averaged {domain} {phase_name}particle "
                    "differential stoichiometry distribution [V-1]"
                ]
                U = variables[
                    f"X-averaged {domain} {phase_name}particle "
                    "potential distribution [V]"
                ]
                # broadcast to "particle size" domain then again into "particle"
                T = pybamm.PrimaryBroadcast(
                    variables[f"X-averaged {domain} electrode temperature [K]"],
                    [f"{domain} {phase_name}particle size"],
                )
                T = pybamm.PrimaryBroadcast(T, [f"{domain} {phase_name}particle"])
                j = variables[
                    f"X-averaged {domain} electrode {phase_name}interfacial "
                    "current density distribution [A.m-2]"
                ]

        # Note: diffusivity is given as a function of concentration here,
        # not stoichiometry
        c_max = self.phase_param.c_max
        current = variables["Total current density [A.m-2]"]
        D_eff = self._get_effective_diffusivity(x * c_max, T, current)
        f = self.param.F / (self.param.R * T)
        N_s = c_max * x * (1 - x) * f * D_eff * pybamm.grad(U)
        variables.update(
            {
                f"{Domain} {phase_name}particle rhs [V.s-1]": -(1 / (R_broad_nondim**2))
                * pybamm.div(N_s)
                / c_max
                / dxdU,
                f"{Domain} {phase_name}particle bc [V.m-1]": j
                * R_nondim
                / self.param.F
                / pybamm.surf(c_max * x * (1 - x) * f * D_eff),
            }
        )

        if self.x_average is True:
            if self.size_distribution is True:
                D_eff = pybamm.TertiaryBroadcast(D_eff, [f"{domain} electrode"])
                N_s = pybamm.TertiaryBroadcast(N_s, [f"{domain} electrode"])
            else:
                D_eff = pybamm.SecondaryBroadcast(D_eff, [f"{domain} electrode"])
                N_s = pybamm.SecondaryBroadcast(N_s, [f"{domain} electrode"])

        if self.size_distribution is True:
            # Size-dependent flux variables
            variables.update(
                self._get_standard_diffusivity_distribution_variables(D_eff)
            )
            variables.update(self._get_standard_flux_distribution_variables(N_s))
            # Size-averaged flux variables
            D_eff = pybamm.size_average(D_eff)
            R = pybamm.SpatialVariable("R", domains=D_eff.domains)
            f_a_dist = self.phase_param.f_a_dist(R)
            N_s = pybamm.Integral(f_a_dist * N_s, R)
        variables.update(self._get_standard_diffusivity_variables(D_eff))
        variables.update(self._get_standard_flux_variables(N_s))

        return variables

    def set_rhs(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        if self.size_distribution is False:
            if self.x_average is False:
                U = variables[f"{Domain} {phase_name}particle potential [V]"]
            else:
                U = variables[f"X-averaged {domain} {phase_name}particle potential [V]"]
        else:
            if self.x_average is False:
                U = variables[
                    f"{Domain} {phase_name}particle potential distribution [V]"
                ]
            else:
                U = variables[
                    f"X-averaged {domain} {phase_name}particle "
                    "potential distribution [V]"
                ]
        self.rhs = {U: variables[f"{Domain} {phase_name}particle rhs [V.s-1]"]}

    def set_boundary_conditions(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        if self.size_distribution is False:
            if self.x_average is False:
                U = variables[f"{Domain} {phase_name}particle potential [V]"]
            else:
                U = variables[f"X-averaged {domain} {phase_name}particle potential [V]"]
        else:
            if self.x_average is False:
                U = variables[
                    f"{Domain} {phase_name}particle potential distribution [V]"
                ]
            else:
                U = variables[
                    f"X-averaged {domain} {phase_name}particle "
                    "potential distribution [V]"
                ]

        rbc = variables[f"{Domain} {phase_name}particle bc [V.m-1]"]
        self.boundary_conditions = {
            U: {"left": (pybamm.Scalar(0), "Neumann"), "right": (rbc, "Neumann")}
        }

    def set_initial_conditions(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        U_init = self.phase_param.U_init
        if self.size_distribution is False:
            if self.x_average is False:
                U = variables[f"{Domain} {phase_name}particle potential [V]"]
            else:
                U = variables[f"X-averaged {domain} {phase_name}particle potential [V]"]
        else:
            if self.x_average is False:
                U = variables[
                    f"{Domain} {phase_name}particle potential distribution [V]"
                ]
            else:
                U = variables[
                    f"X-averaged {domain} {phase_name}particle "
                    "potential distribution [V]"
                ]
        self.initial_conditions = {U: U_init}

    def _get_standard_potential_variables(self, U):
        """
        A private function to obtain the standard variables which can be derived from
        the potential.
        """
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name
        U_surf = pybamm.surf(U)
        U_surf_av = pybamm.x_average(U_surf)
        U_xav = pybamm.x_average(U)
        U_rav = pybamm.r_average(U)
        U_av = pybamm.r_average(U_xav)
        variables = {
            f"{Domain} {phase_name}particle potential [V]": U,
            f"X-averaged {domain} {phase_name}particle potential [V]": U_xav,
            f"R-averaged {domain} {phase_name}particle potential [V]": U_rav,
            f"Average {domain} {phase_name}particle potential [V]": U_av,
            f"{Domain} {phase_name}particle surface potential [V]": U_surf,
            f"X-averaged {domain} {phase_name}particle "
            "surface potential [V]": U_surf_av,
            f"Minimum {domain} {phase_name}particle potential [V]": pybamm.min(U),
            f"Maximum {domain} {phase_name}particle potential [V]": pybamm.max(U),
            f"Minimum {domain} {phase_name}particle surface potential [V]": pybamm.min(
                U_surf
            ),
            f"Maximum {domain} {phase_name}particle surface potential [V]": pybamm.max(
                U_surf
            ),
        }
        return variables

    def _get_standard_potential_distribution_variables(self, U):
        """
        A private function to obtain the standard variables which can be derived from
        the potential distribution in particle size.
        """
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        # Broadcast and x-average when necessary
        if U.domain == [f"{domain} {phase_name}particle size"] and U.domains[
            "secondary"
        ] != [f"{domain} electrode"]:
            # X-avg potential distribution
            U_xav_distribution = pybamm.PrimaryBroadcast(
                U, [f"{domain} {phase_name}particle"]
            )

            # Surface potential distribution variables
            U_surf_xav_distribution = U
            U_surf_distribution = pybamm.SecondaryBroadcast(
                U_surf_xav_distribution, [f"{domain} electrode"]
            )

            # potential distribution in all domains.
            U_distribution = pybamm.PrimaryBroadcast(
                U_surf_distribution, [f"{domain} {phase_name}particle"]
            )
        elif U.domain == [f"{domain} {phase_name}particle"] and (
            U.domains["tertiary"] != [f"{domain} electrode"]
        ):
            # X-avg potential distribution
            U_xav_distribution = U

            # Surface potential distribution variables
            U_surf_xav_distribution = pybamm.surf(U_xav_distribution)
            U_surf_distribution = pybamm.SecondaryBroadcast(
                U_surf_xav_distribution, [f"{domain} electrode"]
            )

            # potential distribution in all domains
            U_distribution = pybamm.TertiaryBroadcast(
                U_xav_distribution, [f"{domain} electrode"]
            )
        elif U.domain == [f"{domain} {phase_name}particle size"] and U.domains[
            "secondary"
        ] == [f"{domain} electrode"]:
            # Surface potential distribution variables
            U_surf_distribution = U
            U_surf_xav_distribution = pybamm.x_average(U)

            # X-avg potential distribution
            U_xav_distribution = pybamm.PrimaryBroadcast(
                U_surf_xav_distribution, [f"{domain} {phase_name}particle"]
            )

            # potential distribution in all domains
            U_distribution = pybamm.PrimaryBroadcast(
                U_surf_distribution, [f"{domain} {phase_name}particle"]
            )
        else:
            U_distribution = U

            # x-average the *tertiary* domain.
            # NOTE: not yet implemented. Make 0.5 everywhere
            U_xav_distribution = pybamm.FullBroadcast(
                0.5,
                [f"{domain} {phase_name}particle"],
                {
                    "secondary": f"{domain} {phase_name}particle size",
                    "tertiary": "current collector",
                },
            )

            # Surface potential distribution variables
            U_surf_distribution = pybamm.surf(U)
            U_surf_xav_distribution = pybamm.x_average(U_surf_distribution)

        U_rav_distribution = pybamm.r_average(U_distribution)
        U_av_distribution = pybamm.x_average(U_rav_distribution)

        variables = {
            f"{Domain} {phase_name}particle potential distribution [V]": U_distribution,
            f"X-averaged {domain} {phase_name}particle potential "
            "distribution [V]": U_xav_distribution,
            f"R-averaged {domain} {phase_name}particle potential "
            "distribution [V]": U_rav_distribution,
            f"Average {domain} {phase_name}particle potential "
            "distribution [V]": U_av_distribution,
            f"{Domain} {phase_name}particle surface potential"
            " distribution [V]": U_surf_distribution,
            f"X-averaged {domain} {phase_name}particle surface potential "
            "distribution [V]": U_surf_xav_distribution,
        }
        return variables


class MSMRStoichiometryVariables(BaseParticle):
    def __init__(self, param, domain, options, phase="primary", x_average=False):
        super().__init__(param, domain, options, phase)
        self.x_average = x_average

    def get_coupled_variables(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        U = variables[f"{Domain} {phase_name}particle potential [V]"]
        T = variables[f"{Domain} electrode temperature [K]"]

        # Standard fractional occupancy variables (these are indexed by reaction number)
        variables.update(self._get_standard_fractional_occupancy_variables(U, T))
        variables.update(
            self._get_standard_differential_fractional_occupancy_variables(U, T)
        )

        # Calculate the (total) stoichiometry from the potential
        x = self.phase_param.x(U, T)
        dxdU = self.phase_param.dxdU(U, T)

        # Standard (total) stoichiometry and concentration variables (size-independent)
        c_max = self.phase_param.c_max
        c_s = x * c_max
        variables.update(self._get_standard_concentration_variables(c_s))
        variables.update(self._get_standard_differential_stoichiometry_variables(dxdU))

        if self.size_distribution is True:
            U_distribution = variables[
                f"{Domain} {phase_name}particle potential distribution [V]"
            ]
            T = variables[f"{Domain} electrode temperature [K]"]
            T_distribution = pybamm.PrimaryBroadcast(
                pybamm.PrimaryBroadcast(T, f"{domain} particle size"),
                f"{domain} particle",
            )
            # Calculate the stoichiometry distribution from the potential distribution
            x_distribution = self.phase_param.x(U_distribution, T_distribution)
            dxdU_distribution = self.phase_param.dxdU(U_distribution, T_distribution)

            # Standard stoichiometry and concentration distribution variables
            # (size-dependent)
            c_s_distribution = x_distribution * c_max
            variables.update(
                self._get_standard_concentration_distribution_variables(
                    c_s_distribution
                )
            )
            variables.update(
                self._get_standard_differential_stoichiometry_distribution_variables(
                    dxdU_distribution
                )
            )
        return variables

    def _get_standard_fractional_occupancy_variables(self, U, T):
        options = self.options
        domain = self.domain
        d = domain[0]
        variables = {}
        # Loop over all reactions
        N = int(getattr(options, domain)["number of MSMR reactions"])
        for i in range(N):
            x = self.phase_param.x_j(U, T, i)
            x_surf = pybamm.surf(x)
            x_surf_av = pybamm.x_average(x_surf)
            x_xav = pybamm.x_average(x)
            x_rav = pybamm.r_average(x)
            x_av = pybamm.r_average(x_xav)
            variables.update(
                {
                    f"x_{d}_{i}": x,
                    f"X-averaged x_{d}_{i}": x_xav,
                    f"R-averaged x_{d}_{i}": x_rav,
                    f"Average x_{d}_{i}": x_av,
                    f"Surface x_{d}_{i}": x_surf,
                    f"X-averaged surface x_{d}_{i}": x_surf_av,
                }
            )
        return variables

    def _get_standard_differential_fractional_occupancy_variables(self, U, T):
        options = self.options
        domain = self.domain
        d = domain[0]
        variables = {}
        # Loop over all reactions
        N = int(getattr(options, domain)["number of MSMR reactions"])
        for i in range(N):
            dxdU = self.phase_param.dxdU_j(U, T, i)
            dxdU_surf = pybamm.surf(dxdU)
            dxdU_surf_av = pybamm.x_average(dxdU_surf)
            dxdU_xav = pybamm.x_average(dxdU)
            dxdU_rav = pybamm.r_average(dxdU)
            dxdU_av = pybamm.r_average(dxdU_xav)
            variables.update(
                {
                    f"dxdU_{d}_{i}": dxdU,
                    f"X-averaged dxdU_{d}_{i}": dxdU_xav,
                    f"R-averaged dxdU_{d}_{i}": dxdU_rav,
                    f"Average dxdU_{d}_{i}": dxdU_av,
                    f"Surface dxdU_{d}_{i}": dxdU_surf,
                    f"X-averaged surface dxdU_{d}_{i}": dxdU_surf_av,
                }
            )
        return variables

    def _get_standard_differential_stoichiometry_variables(self, dxdU):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        dxdU_surf = pybamm.surf(dxdU)
        dxdU_surf_av = pybamm.x_average(dxdU_surf)
        dxdU_xav = pybamm.x_average(dxdU)
        dxdU_rav = pybamm.r_average(dxdU)
        dxdU_av = pybamm.r_average(dxdU_xav)

        variables = {
            f"{Domain} {phase_name}particle differential stoichiometry [V-1]": dxdU,
            f"X-averaged {domain} {phase_name}particle "
            "differential stoichiometry [V-1]": dxdU_xav,
            f"R-averaged {domain} {phase_name}particle "
            "differential stoichiometry [V-1]": dxdU_rav,
            f"Average {domain} {phase_name}particle differential "
            "stoichiometry [V-1]": dxdU_av,
            f"{Domain} {phase_name}particle surface differential "
            "stoichiometry [V-1]": dxdU_surf,
            f"X-averaged {domain} {phase_name}particle "
            "surface differential stoichiometry [V-1]": dxdU_surf_av,
        }

        return variables

    def _get_standard_differential_stoichiometry_distribution_variables(self, dxdU):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        # Broadcast and x-average when necessary
        if dxdU.domain == [f"{domain} {phase_name}particle size"] and dxdU.domains[
            "secondary"
        ] != [f"{domain} electrode"]:
            # X-avg differential stoichiometry distribution
            dxdU_xav_distribution = pybamm.PrimaryBroadcast(
                dxdU, [f"{domain} {phase_name}particle"]
            )

            # Surface differential stoichiometry distribution variables
            dxdU_surf_xav_distribution = dxdU
            dxdU_surf_distribution = pybamm.SecondaryBroadcast(
                dxdU_surf_xav_distribution, [f"{domain} electrode"]
            )

            # Differential stoichiometry distribution in all domains.
            dxdU_distribution = pybamm.PrimaryBroadcast(
                dxdU_surf_distribution, [f"{domain} {phase_name}particle"]
            )
        elif dxdU.domain == [f"{domain} {phase_name}particle"] and (
            dxdU.domains["tertiary"] != [f"{domain} electrode"]
        ):
            # X-avg differential stoichiometry distribution
            dxdU_xav_distribution = dxdU

            # Surface differential stoichiometry distribution variables
            dxdU_surf_xav_distribution = pybamm.surf(dxdU_xav_distribution)
            dxdU_surf_distribution = pybamm.SecondaryBroadcast(
                dxdU_surf_xav_distribution, [f"{domain} electrode"]
            )

            # Differential stoichiometry distribution in all domains.
            dxdU_distribution = pybamm.TertiaryBroadcast(
                dxdU_xav_distribution, [f"{domain} electrode"]
            )
        elif dxdU.domain == [f"{domain} {phase_name}particle size"] and dxdU.domains[
            "secondary"
        ] == [f"{domain} electrode"]:
            # Surface differential stoichiometry distribution variables
            dxdU_surf_distribution = dxdU
            dxdU_surf_xav_distribution = pybamm.x_average(dxdU)

            # X-avg differential stoichiometry distribution
            dxdU_xav_distribution = pybamm.PrimaryBroadcast(
                dxdU_surf_xav_distribution, [f"{domain} {phase_name}particle"]
            )

            # Differential stoichiometry distribution in all domains
            dxdU_distribution = pybamm.PrimaryBroadcast(
                dxdU_surf_distribution, [f"{domain} {phase_name}particle"]
            )
        else:
            dxdU_distribution = dxdU

            # x-average the *tertiary* domain.
            # NOTE: not yet implemented. Make 0.5 everywhere
            dxdU_xav_distribution = pybamm.FullBroadcast(
                0.5,
                [f"{domain} {phase_name}particle"],
                {
                    "secondary": f"{domain} {phase_name}particle size",
                    "tertiary": "current collector",
                },
            )

            # Surface differential stoichiometry distribution variables
            dxdU_surf_distribution = pybamm.surf(dxdU)
            dxdU_surf_xav_distribution = pybamm.x_average(dxdU_surf_distribution)

        dxdU_rav_distribution = pybamm.r_average(dxdU_distribution)
        dxdU_av_distribution = pybamm.x_average(dxdU_rav_distribution)

        variables = {
            f"{Domain} {phase_name}particle differential stoichiometry distribution "
            "[V-1]": dxdU_distribution,
            f"X-averaged {domain} {phase_name}particle differential stoichiometry "
            "distribution [V-1]": dxdU_xav_distribution,
            f"R-averaged {domain} {phase_name}particle differential stoichiometry "
            "distribution [V-1]": dxdU_rav_distribution,
            f"Average {domain} {phase_name}particle differential stoichiometry "
            "distribution [V-1]": dxdU_av_distribution,
            f"{Domain} {phase_name}particle surface differential stoichiometry"
            " distribution [V-1]": dxdU_surf_distribution,
            f"X-averaged {domain} {phase_name}particle surface differential "
            "stoichiometry distribution [V-1]": dxdU_surf_xav_distribution,
        }
        return variables



import pybamm


class OCVElement(pybamm.BaseSubModel):
    """
    Open-circuit Voltage (OCV) element for
    equivalent circuits.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, options=None):
        super().__init__(param)
        self.model_options = options

    def get_fundamental_variables(self):
        soc = pybamm.Variable("SoC")
        ocv = self.param.ocv(soc)
        variables = {"SoC": soc, "Open-circuit voltage [V]": ocv}
        return variables

    def get_coupled_variables(self, variables):
        current = variables["Current [A]"]

        ocv = variables["Open-circuit voltage [V]"]
        T_cell = variables["Cell temperature [degC]"]

        dUdT = self.param.dUdT(ocv, T_cell)

        T_cell_kelvin = variables["Cell temperature [K]"]
        Q_rev = -current * T_cell_kelvin * dUdT

        variables.update(
            {
                "Entropic change [V/K]": dUdT,
                "Reversible heat generation [W]": Q_rev,
            }
        )

        return variables

    def set_rhs(self, variables):
        soc = variables["SoC"]
        current = variables["Current [A]"]
        cell_capacity = self.param.cell_capacity
        self.rhs = {soc: -current / cell_capacity / 3600}

    def set_initial_conditions(self, variables):
        soc = variables["SoC"]
        self.initial_conditions = {soc: self.param.initial_soc}

    def add_events_from(self, variables):
        soc = variables["SoC"]
        self.events += [
            pybamm.Event("Minimum SoC", soc),
            pybamm.Event("Maximum SoC", 1 - soc),
        ]



from __future__ import annotations
from dataclasses import dataclass
from typing import Literal
import numpy.typing as npt
import numpy as np
import pybamm


@dataclass
class ProcessedVariableTimeIntegral:
    method: Literal["discrete", "continuous"]
    initial_condition: npt.NDArray[np.float64] | None | float
    discrete_times: npt.NDArray[np.float64] | None

    @staticmethod
    def from_pybamm_var(
        var: pybamm.DiscreteTimeSum | pybamm.ExplicitTimeIntegral,
    ) -> ProcessedVariableTimeIntegral:
        if isinstance(var, pybamm.DiscreteTimeSum):
            return ProcessedVariableTimeIntegral(
                method="discrete", initial_condition=0.0, discrete_times=var.sum_times
            )
        elif isinstance(var, pybamm.ExplicitTimeIntegral):
            return ProcessedVariableTimeIntegral(
                method="continuous",
                initial_condition=var.initial_condition.evaluate(),
                discrete_times=None,
            )
        else:
            raise ValueError("Unsupported variable type")  # pragma: no cover



from .process_parameter_data import (
    process_1D_data,
    process_2D_data,
    process_2D_data_csv,
    process_3D_data_csv,
)

__all__ = ['base_parameters', 'bpx', 'constants', 'ecm_parameters',
           'electrical_parameters', 'geometric_parameters',
           'lead_acid_parameters', 'lithium_ion_parameters', 'parameter_sets',
           'parameter_values', 'process_parameter_data',
           'size_distribution_parameters', 'thermal_parameters']



#
# Scalar class
#
from __future__ import annotations
import numpy as np
import numpy.typing as npt
import sympy
from typing import Literal

import pybamm
from pybamm.type_definitions import Numeric


class Scalar(pybamm.Symbol):
    """
    A node in the expression tree representing a scalar value.

    Parameters
    ----------
    value : numeric
        the value returned by the node when evaluated
    name : str, optional
        the name of the node. Defaulted to ``str(value)`` if not provided


    """

    def __init__(
        self,
        value: Numeric,
        name: str | None = None,
    ) -> None:
        # set default name if not provided
        self.value = value
        if name is None:
            name = str(self.value)

        super().__init__(name)

    @classmethod
    def _from_json(cls, snippet: dict):
        return cls(snippet["value"], name=snippet["name"])

    def __str__(self):
        return str(self.value)

    @property
    def value(self):
        """The value returned by the node when evaluated."""
        return self._value

    # address numpy 1.25 deprecation warning: array should have ndim=0 before conversion
    @value.setter
    def value(self, value):
        self._value = (
            np.float64(value.item())
            if isinstance(value, np.ndarray)
            else np.float64(value)
        )

    def set_id(self):
        """See :meth:`pybamm.Symbol.set_id()`."""
        # We must include the value in the hash, since different scalars can be
        # indistinguishable by class and name alone
        self._id = hash((self.__class__, str(self.value)))

    def _base_evaluate(
        self,
        t: float | None = None,
        y: npt.NDArray[np.float64] | None = None,
        y_dot: npt.NDArray[np.float64] | None = None,
        inputs: dict | str | None = None,
    ):
        """See :meth:`pybamm.Symbol._base_evaluate()`."""
        return self._value

    def _jac(self, variable: pybamm.Variable) -> pybamm.Scalar:
        """See :meth:`pybamm.Symbol._jac()`."""
        return pybamm.Scalar(0)

    def create_copy(
        self,
        new_children=None,
        perform_simplifications=True,
    ):
        """See :meth:`pybamm.Symbol.new_copy()`."""
        if new_children is not None:
            raise ValueError("Cannot create a copy of a scalar with new children")
        return Scalar(self.value, self.name)

    def is_constant(self) -> Literal[True]:
        """See :meth:`pybamm.Symbol.is_constant()`."""
        return True

    def to_equation(self):
        """Returns the value returned by the node when evaluated."""
        if self.print_name is not None:
            return sympy.Symbol(self.print_name)
        else:
            return self.value

    def to_json(self):
        """
        Method to serialise a Symbol object into JSON.
        """

        json_dict = {"name": self.name, "id": self.id, "value": self.value}

        return json_dict



#
# Function to create battery geometries
#
import pybamm


def battery_geometry(
    include_particles=True,
    options=None,
    form_factor="pouch",
):
    """
    A convenience function to create battery geometries.

    Parameters
    ----------
    include_particles : bool, optional
        Whether to include particle domains. Can be True (default) or False.
    options : dict, optional
        Dictionary of model options. Necessary for "particle-size geometry",
        relevant for lithium-ion chemistries.
    form_factor : str, optional
        The form factor of the cell. Can be "pouch" (default) or "cylindrical".

    Returns
    -------
    :class:`pybamm.Geometry`
        A geometry class for the battery

    """
    if options is None or type(options) == dict:  # noqa: E721
        options = pybamm.BatteryModelOptions(options)
    geo = pybamm.GeometricParameters(options)
    L_n = geo.n.L
    L_s = geo.s.L
    L_n_L_s = L_n + L_s
    # Override print_name
    L_n_L_s.print_name = "L_n + L_s"

    # Set up electrode/separator/electrode geometry
    geometry = {
        "negative electrode": {"x_n": {"min": 0, "max": L_n}},
        "separator": {"x_s": {"min": L_n, "max": L_n_L_s}},
        "positive electrode": {"x_p": {"min": L_n_L_s, "max": geo.L_x}},
    }

    # Add particle domains
    if include_particles is True:
        for domain in ["negative", "positive"]:
            if options.electrode_types[domain] == "porous":
                geo_domain = geo.domain_params[domain]
                d = domain[0]
                geometry.update(
                    {
                        f"{domain} particle": {
                            f"r_{d}": {"min": 0, "max": geo_domain.prim.R_typ}
                        },
                    }
                )
                phases = int(getattr(options, domain)["particle phases"])
                if phases >= 2:
                    geometry.update(
                        {
                            f"{domain} primary particle": {
                                f"r_{d}_prim": {"min": 0, "max": geo_domain.prim.R_typ}
                            },
                            f"{domain} secondary particle": {
                                f"r_{d}_sec": {"min": 0, "max": geo_domain.sec.R_typ}
                            },
                        }
                    )

    # Add particle size domains
    if (
        options is not None
        and options.negative["particle size"] == "distribution"
        and options.electrode_types["negative"] == "porous"
    ):
        R_min_n = geo.n.prim.R_min
        R_max_n = geo.n.prim.R_max
        geometry.update(
            {
                "negative particle size": {"R_n": {"min": R_min_n, "max": R_max_n}},
            }
        )
        phases = int(options.negative["particle phases"])
        if phases >= 2:
            geometry.update(
                {
                    "negative primary particle size": {
                        "R_n_prim": {"min": R_min_n, "max": R_max_n}
                    },
                    "negative secondary particle size": {
                        "R_n_sec": {"min": R_min_n, "max": R_max_n}
                    },
                }
            )
    if (
        options is not None
        and options.positive["particle size"] == "distribution"
        and options.electrode_types["positive"] == "porous"
    ):
        R_min_p = geo.p.prim.R_min
        R_max_p = geo.p.prim.R_max
        geometry.update(
            {
                "positive particle size": {"R_p": {"min": R_min_p, "max": R_max_p}},
            }
        )
        phases = int(options.positive["particle phases"])
        if phases >= 2:
            geometry.update(
                {
                    "positive primary particle size": {
                        "R_p_prim": {"min": R_min_p, "max": R_max_p}
                    },
                    "positive secondary particle size": {
                        "R_p_sec": {"min": R_min_p, "max": R_max_p}
                    },
                }
            )

    # Add current collector domains
    current_collector_dimension = options["dimensionality"]
    if form_factor == "pouch":
        if current_collector_dimension == 0:
            geometry["current collector"] = {"z": {"position": 1}}
        elif current_collector_dimension == 1:
            geometry["current collector"] = {
                "z": {"min": 0, "max": geo.L_z},
                "tabs": {
                    "negative": {"z_centre": geo.n.centre_z_tab},
                    "positive": {"z_centre": geo.p.centre_z_tab},
                },
            }
        elif current_collector_dimension == 2:
            geometry["current collector"] = {
                "y": {"min": 0, "max": geo.L_y},
                "z": {"min": 0, "max": geo.L_z},
                "tabs": {
                    "negative": {
                        "y_centre": geo.n.centre_y_tab,
                        "z_centre": geo.n.centre_z_tab,
                        "width": geo.n.L_tab,
                    },
                    "positive": {
                        "y_centre": geo.p.centre_y_tab,
                        "z_centre": geo.p.centre_z_tab,
                        "width": geo.p.L_tab,
                    },
                },
            }
    elif form_factor == "cylindrical":
        if current_collector_dimension == 0:
            geometry["current collector"] = {"r_macro": {"position": 1}}
        elif current_collector_dimension == 1:
            geometry["current collector"] = {
                "r_macro": {"min": geo.r_inner, "max": 1},
            }
        else:
            raise pybamm.GeometryError(
                f"Invalid current collector dimension '{current_collector_dimension}' (should be 0 or 1 for "
                "a 'cylindrical' battery geometry)"
            )
    else:
        raise pybamm.GeometryError(
            f"Invalid form factor '{form_factor}' (should be 'pouch' or 'cylindrical'"
        )

    return pybamm.Geometry(geometry)



import pybamm

from pybamm.type_definitions import DomainType


class CoupledVariable(pybamm.Symbol):
    """
    A node in the expression tree representing a variable whose equation is set by a different model or submodel.


    Parameters
    ----------
    name : str
        name of the node
    domain : iterable of str
        list of domains that this coupled variable is valid over
    """

    def __init__(
        self,
        name: str,
        domain: DomainType = None,
    ) -> None:
        super().__init__(name, domain=domain)

    def _evaluate_for_shape(self):
        """
        Returns the scalar 'NaN' to represent the shape of a parameter.
        See :meth:`pybamm.Symbol.evaluate_for_shape()`
        """
        return pybamm.evaluate_for_shape_using_domain(self.domains)

    def create_copy(self):
        """Creates a new copy of the coupled variable."""
        new_coupled_variable = CoupledVariable(self.name, self.domain)
        return new_coupled_variable

    @property
    def children(self):
        return self._children

    @children.setter
    def children(self, expr):
        self._children = expr

    def set_coupled_variable(self, symbol, expr):
        """Sets the children of the coupled variable to the expression passed in expr. If the symbol is not the coupled variable, then it searches the children of the symbol for the coupled variable. The coupled variable will be replaced by its first child (symbol.children[0], which should be expr) in the discretisation step."""
        if self == symbol:
            symbol.children = [
                expr,
            ]
        else:
            for child in symbol.children:
                self.set_coupled_variable(child, expr)
        symbol.set_id()



#
# Lead acid base model class
#

import pybamm


class BaseModel(pybamm.BaseBatteryModel):
    """
    Overwrites default parameters from Base Model with default parameters for
    lead-acid models

    Parameters
    ----------
    options : dict-like, optional
        A dictionary of options to be passed to the model. If this is a dict (and not
        a subtype of dict), it will be processed by :class:`pybamm.BatteryModelOptions`
        to ensure that the options are valid. If this is a subtype of dict, it is
        assumed that the options have already been processed and are valid. This allows
        for the use of custom options classes. The default options are given by
        :class:`pybamm.BatteryModelOptions`.
    name : str, optional
        The name of the model. The default is "Unnamed battery model".
    build : bool, optional
        Whether to build the model on instantiation. Default is True. Setting this
        option to False allows users to change any number of the submodels before
        building the complete model (submodels cannot be changed after the model is
        built).
    """

    def __init__(self, options=None, name="Unnamed lead-acid model", build=False):
        options = options or {}
        # Specify that there are no particles in lead-acid
        options["particle shape"] = "no particles"
        super().__init__(options, name)
        self.param = pybamm.LeadAcidParameters()

        self.set_standard_output_variables()

    @property
    def default_parameter_values(self):
        return pybamm.ParameterValues("Sulzer2019")

    @property
    def default_geometry(self):
        return pybamm.battery_geometry(include_particles=False, options=self.options)

    @property
    def default_var_pts(self):
        # Choose points that give uniform grid for the standard parameter values
        return {"x_n": 25, "x_s": 41, "x_p": 34, "y": 10, "z": 10}

    @property
    def default_quick_plot_variables(self):
        return [
            "Interfacial current density [A.m-2]",
            "Electrolyte concentration [mol.m-3]",
            "Current [A]",
            "Porosity",
            "Electrolyte potential [V]",
            "Voltage [V]",
        ]

    def set_soc_variables(self):
        """Set variables relating to the state of charge."""
        # State of Charge defined as function of electrolyte concentration
        z = pybamm.standard_spatial_vars.z
        soc = (
            pybamm.Integral(
                self.variables["X-averaged electrolyte concentration [mol.m-3]"]
                / self.param.c_e_init,
                z,
            )
            * 100
        )
        self.variables.update({"State of Charge": soc, "Depth of Discharge": 100 - soc})

    def set_open_circuit_potential_submodel(self):
        for domain in ["negative", "positive"]:
            self.submodels[f"{domain} open-circuit potential"] = (
                pybamm.open_circuit_potential.SingleOpenCircuitPotential(
                    self.param, domain, "lead-acid main", self.options, "primary"
                )
            )
            self.submodels[f"{domain} oxygen open-circuit potential"] = (
                pybamm.open_circuit_potential.SingleOpenCircuitPotential(
                    self.param, domain, "lead-acid oxygen", self.options, "primary"
                )
            )

    def set_active_material_submodel(self):
        for domain in ["negative", "positive"]:
            self.submodels[f"{domain} active material"] = (
                pybamm.active_material.Constant(
                    self.param, domain, self.options, "primary"
                )
            )

    def set_sei_submodel(self):
        for domain in ["negative", "positive"]:
            self.submodels[f"{domain} sei"] = pybamm.sei.NoSEI(
                self.param, domain, self.options
            )

    def set_lithium_plating_submodel(self):
        for domain in ["negative", "positive"]:
            self.submodels[f"{domain} lithium plating"] = (
                pybamm.lithium_plating.NoPlating(self.param, domain)
            )

    def set_total_interface_submodel(self):
        self.submodels["total interface"] = pybamm.interface.TotalInterfacialCurrent(
            self.param, "lead-acid", self.options
        )



from .base_through_cell_convection import BaseThroughCellModel
from .no_convection import NoConvection
from .explicit_convection import Explicit
from .full_convection import Full

__all__ = ['base_through_cell_convection', 'explicit_convection',
           'full_convection', 'no_convection']



#
# A model to calculate electrode-specific SOH, adapted to a half-cell
#
import pybamm


class ElectrodeSOHHalfCell(pybamm.BaseModel):
    """Model to calculate electrode-specific SOH for a half-cell, adapted from
    :footcite:t:`Mohtat2019`.
    This model is mainly for internal use, to calculate summary variables in a
    simulation.

    .. math::
        V_{max} = U_w(x_{100}),
    .. math::
        V_{min} = U_w(x_{0}),
    .. math::
        x_0 = x_{100} - \\frac{C}{C_w}.

    Subscript w indicates working electrode and subscript c indicates counter electrode.

    """

    def __init__(self, name="ElectrodeSOH model"):
        pybamm.citations.register("Mohtat2019")
        super().__init__(name)
        param = pybamm.LithiumIonParameters({"working electrode": "positive"})

        x_100 = pybamm.Variable("x_100", bounds=(0, 1))
        x_0 = pybamm.Variable("x_0", bounds=(0, 1))
        Q_w = pybamm.InputParameter("Q_w")
        T_ref = param.T_ref
        U_w = param.p.prim.U
        Q = Q_w * (x_0 - x_100)

        V_max = param.ocp_soc_100
        V_min = param.ocp_soc_0

        self.algebraic = {
            x_100: U_w(x_100, T_ref) - V_max,
            x_0: U_w(x_0, T_ref) - V_min,
        }
        self.initial_conditions = {x_100: 0.8, x_0: 0.2}

        self.variables = {
            "x_100": x_100,
            "x_0": x_0,
            "Q": Q,
            "Uw(x_100)": U_w(x_100, T_ref),
            "Uw(x_0)": U_w(x_0, T_ref),
            "Q_w": Q_w,
        }

    @property
    def default_solver(self):
        # Use AlgebraicSolver as CasadiAlgebraicSolver gives unnecessary warnings
        return pybamm.AlgebraicSolver(method="lsq__trf", tol=1e-7)


def get_initial_stoichiometry_half_cell(
    initial_value,
    parameter_values,
    param=None,
    options=None,
    tol=1e-6,
    inputs=None,
    **kwargs,
):
    """
    Calculate initial stoichiometry to start off the simulation at a particular
    state of charge, given voltage limits, open-circuit potential, etc defined by
    parameter_values

    Parameters
    ----------
    initial_value : float
        Target initial value.
        If integer, interpreted as SOC, must be between 0 and 1.
        If string e.g. "4 V", interpreted as voltage,
        must be between V_min and V_max.
    parameter_values : pybamm.ParameterValues
        The parameter values to use in the calculation
    param : :class:`pybamm.LithiumIonParameters`, optional
        The symbolic parameter set to use for the simulation.
        If not provided, the default parameter set will be used.
    tol : float, optional
        The tolerance for the solver used to compute the initial stoichiometries.
        Default is 1e-6.
    inputs : dict, optional
        A dictionary of input parameters passed to the model.
    Returns
    -------
    x
        The initial stoichiometry that give the desired initial state of charge
    """
    param = pybamm.LithiumIonParameters(options)
    x_0, x_100 = get_min_max_stoichiometries(parameter_values, inputs=inputs)

    if isinstance(initial_value, str) and initial_value.endswith("V"):
        V_init = float(initial_value[:-1])
        V_min = parameter_values.evaluate(param.voltage_low_cut)
        V_max = parameter_values.evaluate(param.voltage_high_cut)

        if not V_min - tol <= V_init <= V_max + tol:
            raise ValueError(
                f"Initial voltage {V_init}V is outside the voltage limits "
                f"({V_min}, {V_max})"
            )

        # Solve simple model for initial soc based on target voltage
        soc_model = pybamm.BaseModel()
        soc = pybamm.Variable("soc")
        Up = param.p.prim.U
        T_ref = parameter_values["Reference temperature [K]"]
        x = x_0 + soc * (x_100 - x_0)

        soc_model.algebraic[soc] = Up(x, T_ref) - V_init
        # initial guess for soc linearly interpolates between 0 and 1
        # based on V linearly interpolating between V_max and V_min
        soc_model.initial_conditions[soc] = (V_init - V_min) / (V_max - V_min)
        soc_model.variables["soc"] = soc
        parameter_values.process_model(soc_model)
        initial_soc = (
            pybamm.AlgebraicSolver(tol=tol).solve(soc_model, [0])["soc"].data[0]
        )
    elif isinstance(initial_value, (int, float)):
        initial_soc = initial_value
        if not 0 <= initial_soc <= 1:
            raise ValueError("Initial SOC should be between 0 and 1")

    else:
        raise ValueError(
            "Initial value must be a float between 0 and 1, or a string ending in 'V'"
        )

    x = x_0 + initial_soc * (x_100 - x_0)

    return x


def get_min_max_stoichiometries(parameter_values, options=None, inputs=None):
    """
    Get the minimum and maximum stoichiometries from the parameter values

    Parameters
    ----------
    parameter_values : pybamm.ParameterValues
        The parameter values to use in the calculation
    options : dict, optional
        A dictionary of options to be passed to the parameters, see
        :class:`pybamm.BatteryModelOptions`.
        If None, the default is used: {"working electrode": "positive"}
    """
    inputs = inputs or {}
    if options is None:
        options = {"working electrode": "positive"}
    esoh_model = pybamm.lithium_ion.ElectrodeSOHHalfCell("ElectrodeSOH")
    param = pybamm.LithiumIonParameters(options)
    Q_w = parameter_values.evaluate(param.p.Q_init, inputs=inputs)
    # Add Q_w to input parameters
    all_inputs = {**inputs, "Q_w": Q_w}
    esoh_sim = pybamm.Simulation(esoh_model, parameter_values=parameter_values)
    esoh_sol = esoh_sim.solve([0], inputs=all_inputs)
    x_0, x_100 = esoh_sol["x_0"].data[0], esoh_sol["x_100"].data[0]
    return x_0, x_100



import pybamm
import os
import numpy as np


def graphite_LGM50_electrolyte_exchange_current_density_Chen2020(
    c_e, c_s_surf, c_s_max, T
):
    """
    Exchange-current density for Butler-Volmer reactions between graphite and LiPF6 in
    EC:DMC.

    References
    ----------
    .. [1] Chang-Hui Chen, Ferran Brosa Planella, Kieran O’Regan, Dominika Gastol, W.
    Dhammika Widanage, and Emma Kendrick. "Development of Experimental Techniques for
    Parameterization of Multi-scale Lithium-ion Battery Models." Journal of the
    Electrochemical Society 167 (2020): 080534.

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    c_s_surf : :class:`pybamm.Symbol`
        Particle concentration [mol.m-3]
    c_s_max : :class:`pybamm.Symbol`
        Maximum particle concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]
    """
    m_ref = 6.48e-7  # (A/m2)(m3/mol)**1.5 - includes ref concentrations
    E_r = 35000
    arrhenius = np.exp(E_r / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return m_ref * arrhenius * c_e**0.5 * c_s_surf**0.5 * (c_s_max - c_s_surf) ** 0.5


def silicon_ocp_lithiation_Mark2016(sto):
    """
    silicon Open-circuit Potential (OCP) as a a function of the
    stoichiometry. The fit is taken from the Enertech cell [1], which is only accurate
    for 0 < sto < 1.

    References
    ----------
    .. [1] Verbrugge M, Baker D, Xiao X. Formulation for the treatment of multiple
    electrochemical reactions and associated speciation for the Lithium-Silicon
    electrode[J]. Journal of The Electrochemical Society, 2015, 163(2): A262.

    Parameters
    ----------
    sto: double
       stoichiometry of material (li-fraction)

    Returns
    -------
    :class:`pybamm.Symbol`
        OCP [V]
    """
    p1 = -96.63
    p2 = 372.6
    p3 = -587.6
    p4 = 489.9
    p5 = -232.8
    p6 = 62.99
    p7 = -9.286
    p8 = 0.8633

    U_lithiation = (
        p1 * sto**7
        + p2 * sto**6
        + p3 * sto**5
        + p4 * sto**4
        + p5 * sto**3
        + p6 * sto**2
        + p7 * sto
        + p8
    )
    return U_lithiation


def silicon_ocp_delithiation_Mark2016(sto):
    """
    silicon Open-circuit Potential (OCP) as a a function of the
    stoichiometry. The fit is taken from the Enertech cell [1], which is only accurate
    for 0 < sto < 1.

    References
    ----------
    .. [1] Verbrugge M, Baker D, Xiao X. Formulation for the treatment of multiple
    electrochemical reactions and associated speciation for the Lithium-Silicon
    electrode[J]. Journal of The Electrochemical Society, 2015, 163(2): A262.

    Parameters
    ----------
    sto: double
       stoichiometry of material (li-fraction)

    Returns
    -------
    :class:`pybamm.Symbol`
        OCP [V]
    """
    p1 = -51.02
    p2 = 161.3
    p3 = -205.7
    p4 = 140.2
    p5 = -58.76
    p6 = 16.87
    p7 = -3.792
    p8 = 0.9937

    U_delithiation = (
        p1 * sto**7
        + p2 * sto**6
        + p3 * sto**5
        + p4 * sto**4
        + p5 * sto**3
        + p6 * sto**2
        + p7 * sto
        + p8
    )
    return U_delithiation


def silicon_LGM50_electrolyte_exchange_current_density_Chen2020(
    c_e, c_s_surf, c_s_max, T
):
    """
    Exchange-current density for Butler-Volmer reactions between silicon and LiPF6 in
    EC:DMC.

    References
    ----------
    .. [1] Chang-Hui Chen, Ferran Brosa Planella, Kieran O’Regan, Dominika Gastol, W.
    Dhammika Widanage, and Emma Kendrick. "Development of Experimental Techniques for
    Parameterization of Multi-scale Lithium-ion Battery Models." Journal of the
    Electrochemical Society 167 (2020): 080534.

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    c_s_surf : :class:`pybamm.Symbol`
        Particle concentration [mol.m-3]
    c_s_max : :class:`pybamm.Symbol`
        Maximum particle concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]
    """

    m_ref = (
        6.48e-7 * 28700 / 278000
    )  # (A/m2)(m3/mol)**1.5 - includes ref concentrations
    E_r = 35000
    arrhenius = np.exp(E_r / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return m_ref * arrhenius * c_e**0.5 * c_s_surf**0.5 * (c_s_max - c_s_surf) ** 0.5


def nmc_LGM50_ocp_Chen2020(sto):
    """
    LG M50 NMC open-circuit potential as a function of stoichiometry, fit taken
    from [1].

    References
    ----------
    .. [1] Chang-Hui Chen, Ferran Brosa Planella, Kieran O’Regan, Dominika Gastol, W.
    Dhammika Widanage, and Emma Kendrick. "Development of Experimental Techniques for
    Parameterization of Multi-scale Lithium-ion Battery Models." Journal of the
    Electrochemical Society 167 (2020): 080534.

    Parameters
    ----------
    sto: :class:`pybamm.Symbol`
        Electrode stoichiometry

    Returns
    -------
    :class:`pybamm.Symbol`
        Open-circuit potential
    """

    u_eq = (
        -0.8090 * sto
        + 4.4875
        - 0.0428 * np.tanh(18.5138 * (sto - 0.5542))
        - 17.7326 * np.tanh(15.7890 * (sto - 0.3117))
        + 17.5842 * np.tanh(15.9308 * (sto - 0.3120))
    )

    return u_eq


def nmc_LGM50_electrolyte_exchange_current_density_Chen2020(c_e, c_s_surf, c_s_max, T):
    """
    Exchange-current density for Butler-Volmer reactions between NMC and LiPF6 in
    EC:DMC.

    References
    ----------
    .. [1] Chang-Hui Chen, Ferran Brosa Planella, Kieran O’Regan, Dominika Gastol, W.
    Dhammika Widanage, and Emma Kendrick. "Development of Experimental Techniques for
    Parameterization of Multi-scale Lithium-ion Battery Models." Journal of the
    Electrochemical Society 167 (2020): 080534.

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    c_s_surf : :class:`pybamm.Symbol`
        Particle concentration [mol.m-3]
    c_s_max : :class:`pybamm.Symbol`
        Maximum particle concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]
    """
    m_ref = 3.42e-6  # (A/m2)(m3/mol)**1.5 - includes ref concentrations
    E_r = 17800
    arrhenius = np.exp(E_r / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return m_ref * arrhenius * c_e**0.5 * c_s_surf**0.5 * (c_s_max - c_s_surf) ** 0.5


def electrolyte_diffusivity_Nyman2008(c_e, T):
    """
    Diffusivity of LiPF6 in EC:EMC (3:7) as a function of ion concentration. The data
    comes from [1]

    References
    ----------
    .. [1] A. Nyman, M. Behm, and G. Lindbergh, "Electrochemical characterisation and
    modelling of the mass transport phenomena in LiPF6-EC-EMC electrolyte,"
    Electrochim. Acta, vol. 53, no. 22, pp. 6356–6365, 2008.

    Parameters
    ----------
    c_e: :class:`pybamm.Symbol`
        Dimensional electrolyte concentration
    T: :class:`pybamm.Symbol`
        Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """

    D_c_e = 8.794e-11 * (c_e / 1000) ** 2 - 3.972e-10 * (c_e / 1000) + 4.862e-10

    # Nyman et al. (2008) does not provide temperature dependence

    return D_c_e


def electrolyte_conductivity_Nyman2008(c_e, T):
    """
    Conductivity of LiPF6 in EC:EMC (3:7) as a function of ion concentration. The data
    comes from [1].

    References
    ----------
    .. [1] A. Nyman, M. Behm, and G. Lindbergh, "Electrochemical characterisation and
    modelling of the mass transport phenomena in LiPF6-EC-EMC electrolyte,"
    Electrochim. Acta, vol. 53, no. 22, pp. 6356–6365, 2008.

    Parameters
    ----------
    c_e: :class:`pybamm.Symbol`
        Dimensional electrolyte concentration
    T: :class:`pybamm.Symbol`
        Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """

    sigma_e = (
        0.1297 * (c_e / 1000) ** 3 - 2.51 * (c_e / 1000) ** 1.5 + 3.329 * (c_e / 1000)
    )

    # Nyman et al. (2008) does not provide temperature dependence

    return sigma_e


# Load data in the appropriate format
path, _ = os.path.split(os.path.abspath(__file__))
graphite_ocp_Enertech_Ai2020_data = pybamm.parameters.process_1D_data(
    "graphite_ocp_Enertech_Ai2020.csv", path=path
)


def graphite_ocp_Enertech_Ai2020(sto):
    name, (x, y) = graphite_ocp_Enertech_Ai2020_data
    return pybamm.Interpolant(x, y, sto, name=name, interpolator="cubic")


# Call dict via a function to avoid errors when editing in place
def get_parameter_values():
    """
    Parameters for a composite graphite/silicon negative electrode, from the paper
    :footcite:t:`Ai2022`, based on the paper :footcite:t:`Chen2020`, and references
    therein.

    SEI parameters are example parameters for composite SEI on silicon/graphite. Both
    phases use the same values, from the paper :footcite:t:`Yang2017`
    """

    return {
        "chemistry": "lithium_ion",
        # sei
        "Primary: Ratio of lithium moles to SEI moles": 2.0,
        "Primary: SEI partial molar volume [m3.mol-1]": 9.585e-05,
        "Primary: SEI reaction exchange current density [A.m-2]": 1.5e-07,
        "Primary: SEI resistivity [Ohm.m]": 200000.0,
        "Primary: SEI solvent diffusivity [m2.s-1]": 2.5e-22,
        "Primary: Bulk solvent concentration [mol.m-3]": 2636.0,
        "Primary: SEI open-circuit potential [V]": 0.4,
        "Primary: SEI electron conductivity [S.m-1]": 8.95e-14,
        "Primary: SEI lithium interstitial diffusivity [m2.s-1]": 1e-20,
        "Primary: Lithium interstitial reference concentration [mol.m-3]": 15.0,
        "Primary: Initial SEI thickness [m]": 5e-09,
        "Primary: EC initial concentration in electrolyte [mol.m-3]": 4541.0,
        "Primary: EC diffusivity [m2.s-1]": 2e-18,
        "Primary: SEI kinetic rate constant [m.s-1]": 1e-12,
        "Primary: SEI growth activation energy [J.mol-1]": 0.0,
        "Secondary: Ratio of lithium moles to SEI moles": 2.0,
        "Secondary: SEI partial molar volume [m3.mol-1]": 9.585e-05,
        "Secondary: SEI reaction exchange current density [A.m-2]": 1.5e-07,
        "Secondary: SEI resistivity [Ohm.m]": 200000.0,
        "Secondary: SEI solvent diffusivity [m2.s-1]": 2.5e-22,
        "Secondary: Bulk solvent concentration [mol.m-3]": 2636.0,
        "Secondary: SEI open-circuit potential [V]": 0.4,
        "Secondary: SEI electron conductivity [S.m-1]": 8.95e-14,
        "Secondary: SEI lithium interstitial diffusivity [m2.s-1]": 1e-20,
        "Secondary: Lithium interstitial reference concentration [mol.m-3]": 15.0,
        "Secondary: Initial SEI thickness [m]": 5e-09,
        "Secondary: EC initial concentration in electrolyte [mol.m-3]": 4541.0,
        "Secondary: EC diffusivity [m2.s-1]": 2e-18,
        "Secondary: SEI kinetic rate constant [m.s-1]": 1e-12,
        "Secondary: SEI growth activation energy [J.mol-1]": 0.0,
        "Positive electrode reaction-driven LAM factor [m3.mol-1]": 0.0,
        # cell
        "Negative current collector thickness [m]": 1.2e-05,
        "Negative electrode thickness [m]": 8.52e-05,
        "Separator thickness [m]": 1.2e-05,
        "Positive electrode thickness [m]": 7.56e-05,
        "Positive current collector thickness [m]": 1.6e-05,
        "Electrode height [m]": 0.065,
        "Electrode width [m]": 1.58,
        "Cell cooling surface area [m2]": 0.00531,
        "Cell volume [m3]": 2.42e-05,
        "Cell thermal expansion coefficient [m.K-1]": 1.1e-06,
        "Negative current collector conductivity [S.m-1]": 58411000.0,
        "Positive current collector conductivity [S.m-1]": 36914000.0,
        "Negative current collector density [kg.m-3]": 8960.0,
        "Positive current collector density [kg.m-3]": 2700.0,
        "Negative current collector specific heat capacity [J.kg-1.K-1]": 385.0,
        "Positive current collector specific heat capacity [J.kg-1.K-1]": 897.0,
        "Negative current collector thermal conductivity [W.m-1.K-1]": 401.0,
        "Positive current collector thermal conductivity [W.m-1.K-1]": 237.0,
        "Nominal cell capacity [A.h]": 5.0,
        "Current function [A]": 5.0,
        "Contact resistance [Ohm]": 0,
        # negative electrode
        "Negative electrode conductivity [S.m-1]": 215.0,
        "Primary: Maximum concentration in negative electrode [mol.m-3]": 28700.0,
        "Primary: Initial concentration in negative electrode [mol.m-3]": 27700.0,
        "Primary: Negative particle diffusivity [m2.s-1]": 5.5e-14,
        "Primary: Negative electrode OCP [V]": graphite_ocp_Enertech_Ai2020,
        "Negative electrode porosity": 0.25,
        "Primary: Negative electrode active material volume fraction": 0.735,
        "Primary: Negative particle radius [m]": 5.86e-06,
        "Negative electrode Bruggeman coefficient (electrolyte)": 1.5,
        "Negative electrode Bruggeman coefficient (electrode)": 0,
        "Negative electrode charge transfer coefficient": 0.5,
        "Negative electrode double-layer capacity [F.m-2]": 0.2,
        "Primary: Negative electrode exchange-current density [A.m-2]"
        "": graphite_LGM50_electrolyte_exchange_current_density_Chen2020,
        "Primary: Negative electrode density [kg.m-3]": 1657.0,
        "Negative electrode specific heat capacity [J.kg-1.K-1]": 700.0,
        "Negative electrode thermal conductivity [W.m-1.K-1]": 1.7,
        "Primary: Negative electrode OCP entropic change [V.K-1]": 0.0,
        "Secondary: Maximum concentration in negative electrode [mol.m-3]": 278000.0,
        "Secondary: Initial concentration in negative electrode [mol.m-3]": 276610.0,
        "Secondary: Negative particle diffusivity [m2.s-1]": 1.67e-14,
        "Secondary: Negative electrode lithiation OCP [V]"
        "": silicon_ocp_lithiation_Mark2016,
        "Secondary: Negative electrode delithiation OCP [V]"
        "": silicon_ocp_delithiation_Mark2016,
        "Secondary: Negative electrode active material volume fraction": 0.015,
        "Secondary: Negative particle radius [m]": 1.52e-06,
        "Secondary: Negative electrode exchange-current density [A.m-2]"
        "": silicon_LGM50_electrolyte_exchange_current_density_Chen2020,
        "Secondary: Negative electrode density [kg.m-3]": 2650.0,
        "Secondary: Negative electrode OCP entropic change [V.K-1]": 0.0,
        # positive electrode
        "Positive electrode conductivity [S.m-1]": 0.18,
        "Maximum concentration in positive electrode [mol.m-3]": 63104.0,
        "Positive particle diffusivity [m2.s-1]": 4e-15,
        "Positive electrode OCP [V]": nmc_LGM50_ocp_Chen2020,
        "Positive electrode porosity": 0.335,
        "Positive electrode active material volume fraction": 0.665,
        "Positive particle radius [m]": 5.22e-06,
        "Positive electrode Bruggeman coefficient (electrolyte)": 1.5,
        "Positive electrode Bruggeman coefficient (electrode)": 0,
        "Positive electrode charge transfer coefficient": 0.5,
        "Positive electrode double-layer capacity [F.m-2]": 0.2,
        "Positive electrode exchange-current density [A.m-2]"
        "": nmc_LGM50_electrolyte_exchange_current_density_Chen2020,
        "Positive electrode density [kg.m-3]": 3262.0,
        "Positive electrode specific heat capacity [J.kg-1.K-1]": 700.0,
        "Positive electrode thermal conductivity [W.m-1.K-1]": 2.1,
        "Positive electrode OCP entropic change [V.K-1]": 0.0,
        # separator
        "Separator porosity": 0.47,
        "Separator Bruggeman coefficient (electrolyte)": 1.5,
        "Separator density [kg.m-3]": 397.0,
        "Separator specific heat capacity [J.kg-1.K-1]": 700.0,
        "Separator thermal conductivity [W.m-1.K-1]": 0.16,
        # electrolyte
        "Initial concentration in electrolyte [mol.m-3]": 1000.0,
        "Cation transference number": 0.2594,
        "Thermodynamic factor": 1.0,
        "Electrolyte diffusivity [m2.s-1]": electrolyte_diffusivity_Nyman2008,
        "Electrolyte conductivity [S.m-1]": electrolyte_conductivity_Nyman2008,
        # experiment
        "Reference temperature [K]": 298.15,
        "Total heat transfer coefficient [W.m-2.K-1]": 10.0,
        "Ambient temperature [K]": 298.15,
        "Number of electrodes connected in parallel to make a cell": 1.0,
        "Number of cells connected in series to make a battery": 1.0,
        "Lower voltage cut-off [V]": 2.5,
        "Upper voltage cut-off [V]": 4.2,
        "Open-circuit voltage at 0% SOC [V]": 2.5,
        "Open-circuit voltage at 100% SOC [V]": 4.2,
        "Initial concentration in negative electrode [mol.m-3]": 29866.0,
        "Initial concentration in positive electrode [mol.m-3]": 17038.0,
        "Initial temperature [K]": 298.15,
        # citations
        "citations": ["Chen2020", "Ai2022"],
    }



import pybamm
import numpy as np


def lead_ocp_Bode1977(m):
    """
    Dimensional open-circuit voltage in the negative (lead) electrode [V], from [1]_,
    as a function of the molar mass m [mol.kg-1].

    References
    ----------
    .. [1] H Bode. Lead-acid batteries. John Wiley and Sons, Inc., New York, NY, 1977.

    """
    U = (
        -0.294
        - 0.074 * np.log10(m)
        - 0.030 * np.log10(m) ** 2
        - 0.031 * np.log10(m) ** 3
        - 0.012 * np.log10(m) ** 4
    )
    return U


def lead_exchange_current_density_Sulzer2019(c_e, T):
    """
    Dimensional exchange-current density in the negative (lead) electrode, from [1]_

    References
    ----------
    .. [1] V. Sulzer, S. J. Chapman, C. P. Please, D. A. Howey, and C. W. Monroe,
    “Faster lead-acid battery simulations from porous-electrode theory: Part I. Physical
    model.”
    [Journal of the Electrochemical Society](https://doi.org/10.1149/2.0301910jes),
    166(12), 2363 (2019).

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]

    """
    j0_ref = 0.06  # srinivasan2003mathematical
    c_e_init = pybamm.Parameter("Initial concentration in electrolyte [mol.m-3]")
    j0 = j0_ref * (c_e / c_e_init)

    return j0


def lead_dioxide_ocp_Bode1977(m):
    """
    Dimensional open-circuit voltage in the positive (lead-dioxide) electrode [V],
    from [1]_, as a function of the molar mass m [mol.kg-1].

    References
    ----------
    .. [1] H Bode. Lead-acid batteries. John Wiley and Sons, Inc., New York, NY, 1977.

    """
    U = (
        1.628
        + 0.074 * np.log10(m)
        + 0.033 * np.log10(m) ** 2
        + 0.043 * np.log10(m) ** 3
        + 0.022 * np.log10(m) ** 4
    )
    return U


def lead_dioxide_exchange_current_density_Sulzer2019(c_e, T):
    """
    Dimensional exchange-current density in the positive electrode, from [1]_

    References
    ----------
    .. [1] V. Sulzer, S. J. Chapman, C. P. Please, D. A. Howey, and C. W. Monroe,
    “Faster lead-acid battery simulations from porous-electrode theory: Part I. Physical
    model.”
    [Journal of the Electrochemical Society](https://doi.org/10.1149/2.0301910jes),
    166(12), 2363 (2019).

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]

    """
    c_ox = 0
    c_hy = 0
    param = pybamm.LeadAcidParameters()
    c_w_dim = (1 - c_e * param.V_e - c_ox * param.V_ox - c_hy * param.V_hy) / param.V_w
    c_w_ref = (1 - param.c_e_init * param.V_e) / param.V_w
    c_w = c_w_dim / c_w_ref

    j0_ref = 0.004  # srinivasan2003mathematical
    j0 = j0_ref * (c_e / param.c_e_init) ** 2 * c_w

    return j0


def oxygen_exchange_current_density_Sulzer2019(c_e, T):
    """
    Dimensional oxygen exchange-current density in the positive electrode, from [1]_

    References
    ----------
    .. [1] Valentin Sulzer, S. Jon Chapman, Colin P. Please, David A. Howey, and Charles
        W. Monroe. Faster Lead-Acid Battery Simulations from Porous-Electrode Theory:
        Part I. Physical Model. Journal of The Electrochemical Society,
        166(12):A2363-A2371, 2019. doi:10.1149/2.0301910jes.

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]

    """
    j0_ref = 2.5e-23  # srinivasan2003mathematical
    c_e_init = pybamm.Parameter("Initial concentration in electrolyte [mol.m-3]")
    j0 = j0_ref * (c_e / c_e_init)

    return j0


def conductivity_Gu1997(c_e):
    """
    Dimensional conductivity of sulfuric acid [S.m-1], from [1]_ citing [2]_ and
    agreeing with data in [3]_, as a function of the electrolyte concentration
    c_e [mol.m-3].

    References
    ----------
    .. [1] WB Gu, CY Wang, and BY Liaw. Numerical modeling of coupled electrochemical
           and transport processes in lead-acid batteries. Journal of The
           Electrochemical Society, 144(6):2053–2061, 1997.
    .. [2] WH Tiedemann and J Newman. Battery design and optimization. Journal of
           Electrochemical Society, Softbound Proceeding Series, Princeton, New York,
           79(1):23, 1979.
    .. [3] TW Chapman and J Newman. Compilation of selected thermodynamic and transport
           properties of binary electrolytes in aqueous solution. Technical report,
           California Univ., Berkeley. Lawrence Radiation Lab., 1968.

    """
    return c_e * np.exp(6.23 - 1.34e-4 * c_e - 1.61e-8 * c_e**2) * 1e-4


def darken_thermodynamic_factor_Chapman1968(c_e):
    """
    Dimensional Darken thermodynamic factor of sulfuric acid, from data in
    [1, 2]_, as a function of the electrolyte concentration c_e [mol.m-3].

    References
    ----------
    .. [1] TW Chapman and J Newman. Compilation of selected thermodynamic and transport
           properties of binary electrolytes in aqueous solution. Technical report,
           California Univ., Berkeley. Lawrence Radiation Lab., 1968.
    .. [2] KS Pitzer, RN Roy, and LF Silvester. Thermodynamics of electrolytes. 7.
           sulfuric acid. Journal of the American Chemical Society, 99(15):4930–4936,
           1977.

    """
    return 0.49 + 4.1e-4 * c_e


def diffusivity_Gu1997(c_e):
    """
    Dimensional Fickian diffusivity of sulfuric acid [m2.s-1], from [1]_ citing [2]_
    and agreeing with data in [3]_, as a function of the electrolyte concentration
    c_e [mol.m-3].

    References
    ----------
    .. [1] WB Gu, CY Wang, and BY Liaw. Numerical modeling of coupled electrochemical
           and transport processes in lead-acid batteries. Journal of The
           Electrochemical Society, 144(6):2053–2061, 1997.
    .. [2] WH Tiedemann and J Newman. Battery design and optimization. Journal of
           Electrochemical Society, Softbound Proceeding Series, Princeton, New York,
           79(1):23, 1979.
    .. [3] TW Chapman and J Newman. Compilation of selected thermodynamic and transport
           properties of binary electrolytes in aqueous solution. Technical report,
           California Univ., Berkeley. Lawrence Radiation Lab., 1968.

    """
    return (1.75 + 260e-6 * c_e) * 1e-9


# Call dict via a function to avoid errors when editing in place
def get_parameter_values():
    """
    Parameters for BBOXX lead-acid cells, from the paper
    :footcite:t:`sulzer2019asymptotic` and references therein.
    """

    return {
        "chemistry": "lead_acid",
        # cell
        "Negative current collector thickness [m]": 0.0,
        "Negative electrode thickness [m]": 0.0009,
        "Separator thickness [m]": 0.0015,
        "Positive electrode thickness [m]": 0.00125,
        "Positive current collector thickness [m]": 0.0,
        "Electrode height [m]": 0.114,
        "Electrode width [m]": 0.065,
        "Negative tab width [m]": 0.04,
        "Negative tab centre y-coordinate [m]": 0.06,
        "Negative tab centre z-coordinate [m]": 0.114,
        "Positive tab width [m]": 0.04,
        "Positive tab centre y-coordinate [m]": 0.147,
        "Positive tab centre z-coordinate [m]": 0.114,
        "Cell cooling surface area [m2]": 0.154,
        "Cell volume [m3]": 0.00027,
        "Nominal cell capacity [A.h]": 17.0,
        "Current function [A]": 1.0,
        "Negative current collector density [kg.m-3]": 11300.0,
        "Positive current collector density [kg.m-3]": 9375.0,
        "Negative current collector specific heat capacity [J.kg-1.K-1]": 130.0,
        "Positive current collector specific heat capacity [J.kg-1.K-1]": 256.0,
        "Negative current collector thermal conductivity [W.m-1.K-1]": 35.0,
        "Positive current collector thermal conductivity [W.m-1.K-1]": 35.0,
        "Contact resistance [Ohm]": 0,
        # negative electrode
        "Negative electrode conductivity [S.m-1]": 4800000.0,
        "Negative electrode pore size [m]": 1e-07,
        "Maximum porosity of negative electrode": 0.53,
        "Molar volume of lead [m3.mol-1]": 1.82539682539683e-05,
        "Negative electrode volumetric capacity [C.m-3]": 3473000000.0,
        "Negative electrode open-circuit potential [V]": lead_ocp_Bode1977,
        "Negative electrode surface area to volume ratio [m-1]": 2300000.0,
        "Negative electrode Bruggeman coefficient (electrolyte)": 1.5,
        "Negative electrode Bruggeman coefficient (electrode)": 1.5,
        "Negative electrode morphological parameter": 0.6,
        "Negative electrode capacity [C.m-3]": 3473000000.0,
        "Negative electrode exchange-current density [A.m-2]"
        "": lead_exchange_current_density_Sulzer2019,
        "Signed stoichiometry of cations (oxygen reaction)": 4.0,
        "Signed stoichiometry of water (oxygen reaction)": -1.0,
        "Signed stoichiometry of oxygen (oxygen reaction)": 1.0,
        "Electrons in oxygen reaction": 4.0,
        "Negative electrode reference exchange-current density (oxygen) [A.m-2]"
        "": 2.5e-32,
        "Reference oxygen molecule concentration [mol.m-3]": 1000.0,
        "Oxygen reference OCP vs SHE [V]": 1.229,
        "Signed stoichiometry of cations (hydrogen reaction)": 2.0,
        "Signed stoichiometry of hydrogen (hydrogen reaction)": -1.0,
        "Electrons in hydrogen reaction": 2.0,
        "Negative electrode reference exchange-current density (hydrogen) [A.m-2]"
        "": 1.56e-11,
        "Hydrogen reference OCP vs SHE [V]": 0.0,
        "Negative electrode double-layer capacity [F.m-2]": 0.2,
        "Negative electrode density [kg.m-3]": 11300.0,
        "Negative electrode specific heat capacity [J.kg-1.K-1]": 130.0,
        "Negative electrode thermal conductivity [W.m-1.K-1]": 35.0,
        # positive electrode
        "Positive electrode conductivity [S.m-1]": 80000.0,
        "Positive electrode pore size [m]": 1e-07,
        "Maximum porosity of positive electrode": 0.57,
        "Molar volume of lead-dioxide [m3.mol-1]": 2.54797441364606e-05,
        "Molar volume of lead sulfate [m3.mol-1]": 4.81717011128776e-05,
        "Positive electrode volumetric capacity [C.m-3]": 2745000000.0,
        "Positive electrode open-circuit potential [V]": lead_dioxide_ocp_Bode1977,
        "Positive electrode surface area to volume ratio [m-1]": 23000000.0,
        "Positive electrode Bruggeman coefficient (electrolyte)": 1.5,
        "Positive electrode Bruggeman coefficient (electrode)": 1.5,
        "Positive electrode morphological parameter": 0.6,
        "Positive electrode capacity [C.m-3]": 2745000000.0,
        "Positive electrode exchange-current density [A.m-2]"
        "": lead_dioxide_exchange_current_density_Sulzer2019,
        "Positive electrode oxygen exchange-current density [A.m-2]"
        "": oxygen_exchange_current_density_Sulzer2019,
        "Positive electrode Butler-Volmer transfer coefficient": 0.5,
        "Positive electrode reference exchange-current density (hydrogen) [A.m-2]": 0.0,
        "Positive electrode double-layer capacity [F.m-2]": 0.2,
        "Positive electrode density [kg.m-3]": 9375.0,
        "Positive electrode specific heat capacity [J.kg-1.K-1]": 256.0,
        "Positive electrode thermal conductivity [W.m-1.K-1]": 35.0,
        # separator
        "Maximum porosity of separator": 0.92,
        "Separator Bruggeman coefficient (electrolyte)": 1.5,
        "Separator density [kg.m-3]": 1680.0,
        "Separator specific heat capacity [J.kg-1.K-1]": 700.0,
        "Separator thermal conductivity [W.m-1.K-1]": 0.04,
        # electrolyte
        "Initial concentration in electrolyte [mol.m-3]": 5650.0,
        "Cation transference number": 0.7,
        "Thermodynamic factor": 1.0,
        "Partial molar volume of water [m3.mol-1]": 1.75e-05,
        "Partial molar volume of anions [m3.mol-1]": 3.15e-05,
        "Partial molar volume of cations [m3.mol-1]": 1.35e-05,
        "Cation stoichiometry": 1.0,
        "Anion stoichiometry": 1.0,
        "Molar mass of water [kg.mol-1]": 0.01801,
        "Molar mass of cations [kg.mol-1]": 0.001,
        "Molar mass of anions [kg.mol-1]": 0.097,
        "Volume change factor": 1.0,
        "Electrolyte conductivity [S.m-1]": conductivity_Gu1997,
        "Darken thermodynamic factor": darken_thermodynamic_factor_Chapman1968,
        "Electrolyte diffusivity [m2.s-1]": diffusivity_Gu1997,
        "Oxygen diffusivity [m2.s-1]": 2.1e-09,
        "Typical oxygen concentration [mol.m-3]": 1000.0,
        "Hydrogen diffusivity [m2.s-1]": 4.5e-09,
        "Partial molar volume of oxygen molecules [m3.mol-1]": 3.21e-05,
        "Partial molar volume of hydrogen molecules [m3.mol-1]": 2.31e-05,
        "Molar mass of oxygen molecules [kg.mol-1]": 0.032,
        "Molar mass of hydrogen molecules [kg.mol-1]": 0.002,
        # experiment
        "Reference temperature [K]": 294.85,
        "Maximum temperature [K]": 333.15,
        "Ambient temperature [K]": 294.85,
        "Negative current collector surface heat transfer coefficient [W.m-2.K-1]": 0.0,
        "Positive current collector surface heat transfer coefficient [W.m-2.K-1]": 0.0,
        "Negative tab heat transfer coefficient [W.m-2.K-1]": 10.0,
        "Positive tab heat transfer coefficient [W.m-2.K-1]": 10.0,
        "Total heat transfer coefficient [W.m-2.K-1]": 10.0,
        "Edge heat transfer coefficient [W.m-2.K-1]": 0.3,
        "Number of electrodes connected in parallel to make a cell": 8.0,
        "Number of cells connected in series to make a battery": 6.0,
        "Lower voltage cut-off [V]": 1.75,
        "Upper voltage cut-off [V]": 2.42,
        "Initial State of Charge": 1.0,
        "Initial oxygen concentration [mol.m-3]": 0.0,
        "Initial temperature [K]": 294.85,
        # citations
        "citations": ["Sulzer2019physical"],
    }



from __future__ import annotations

from typing import Union
from typing_extensions import TypeAlias
import numpy as np
import numpy.typing as npt
import pybamm
from typing import Any

# numbers.Number should not be used for type hints
Numeric: TypeAlias = Union[int, float, np.number]

# expression tree
ChildValue: TypeAlias = Union[float, npt.NDArray[Any]]
ChildSymbol: TypeAlias = Union[float, npt.NDArray[Any], pybamm.Symbol]

DomainType: TypeAlias = Union[list[str], str, None]
AuxiliaryDomainType: TypeAlias = Union[dict[str, str], None]
DomainsType: TypeAlias = Union[dict[str, Union[list[str], str]], None]



import casadi
import pybamm
import numpy as np
import numbers


class CasadiAlgebraicSolver(pybamm.BaseSolver):
    """Solve a discretised model which contains only (time independent) algebraic
    equations using CasADi's root finding algorithm.

    Note: this solver could be extended for quasi-static models, or models in
    which the time derivative is manually discretised and results in a (possibly
    nonlinear) algebraic system at each time level.

    Parameters
    ----------
    tol : float, optional
        The tolerance for the maximum residual error (default is 1e-6).
    step_tol : float, optional
        The tolerance for the maximum step size (default is 1e-4).
    extra_options : dict, optional
        Any options to pass to the CasADi rootfinder.
        Please consult `CasADi documentation <https://web.casadi.org/python-api/#rootfinding>`_ for
        details. By default:

        .. code-block:: python

            extra_options = {
                # Whether to throw an error if the solver fails to converge.
                "error_on_fail": False,
                # Verbosity level
                "verbose": False,
                # Whether to show warnings when evaluating the model
                "show_eval_warnings": False,
            }
    """

    def __init__(self, tol=1e-6, step_tol=1e-4, extra_options=None):
        super().__init__()
        default_extra_options = {
            "error_on_fail": False,
            "verbose": False,
            "show_eval_warnings": False,
        }
        extra_options = extra_options or {}
        self.extra_options = extra_options | default_extra_options
        self.step_tol = step_tol
        self.tol = tol
        self.name = "CasADi algebraic solver"
        self._algebraic_solver = True
        pybamm.citations.register("Andersson2019")

    @property
    def tol(self):
        return self._tol

    @tol.setter
    def tol(self, value):
        self._tol = value

    @property
    def step_tol(self):
        return self._step_tol

    @step_tol.setter
    def step_tol(self, value):
        self._step_tol = value

    def set_up_root_solver(self, model, inputs_dict, t_eval):
        """Create and return a CasADi rootfinder object. The parameter argument to the
        rootfinder is the concatenated time, differential states, and flattened inputs.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        inputs_dict : dict
            Dictionary of inputs.
        t_eval : :class:`numpy.array`, size (k,)
            The times at which to compute the solution (not used).

        Returns
        -------
        None
            The rootfinder function is stored in the model as `algebraic_root_solver`.
        """
        pybamm.logger.info(f"Start building {self.name}")
        y0 = model.y0

        # The casadi algebraic solver can read rhs equations, but leaves them unchanged
        # i.e. the part of the solution vector that corresponds to the differential
        # equations will be equal to the initial condition provided. This allows this
        # solver to be used for initialising the DAE solvers
        if model.rhs == {}:
            len_rhs = 0
        elif model.len_rhs_and_alg == y0.shape[0]:
            len_rhs = model.len_rhs
        else:
            len_rhs = model.len_rhs + model.len_rhs_sens

        len_alg = y0.shape[0] - len_rhs

        # Set up symbolic variables
        t_sym = casadi.MX.sym("t")
        y_diff_sym = casadi.MX.sym("y_diff", len_rhs)
        y_alg_sym = casadi.MX.sym("y_alg", len_alg)
        y_sym = casadi.vertcat(y_diff_sym, y_alg_sym)

        # Create parameter dictionary
        inputs_casadi = {}
        for name, value in inputs_dict.items():
            if isinstance(value, numbers.Number):
                inputs_casadi[name] = casadi.MX.sym(name)
            else:
                inputs_casadi[name] = casadi.MX.sym(name, value.shape[0])
        inputs_sym = casadi.vertcat(*[p for p in inputs_casadi.values()])

        p_sym = casadi.vertcat(
            t_sym,
            y_diff_sym,
            inputs_sym,
        )

        alg = model.casadi_algebraic(t_sym, y_sym, inputs_sym)

        # Set constraints vector in the casadi format
        # Constrain the unknowns. 0 (default): no constraint on ui, 1: ui >= 0.0,
        # -1: ui <= 0.0, 2: ui > 0.0, -2: ui < 0.0.
        model_alg_lb = model.bounds[0][len_rhs:]
        model_alg_ub = model.bounds[1][len_rhs:]
        constraints = np.zeros_like(model_alg_lb, dtype=int)
        # If the lower bound is positive then the variable must always be positive
        constraints[model_alg_lb >= 0] = 1
        # If the upper bound is negative then the variable must always be negative
        constraints[model_alg_ub <= 0] = -1

        # Set up rootfinder
        model.algebraic_root_solver = casadi.rootfinder(
            "roots",
            "newton",
            dict(x=y_alg_sym, p=p_sym, g=alg),
            {
                **self.extra_options,
                "abstol": self.tol,
                "abstolStep": self.step_tol,
                "constraints": list(constraints),
            },
        )

        pybamm.logger.info(f"Finish building {self.name}")

    def _integrate(self, model, t_eval, inputs_dict=None, t_interp=None):
        """
        Calculate the solution of the algebraic equations through root-finding

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        t_eval : :class:`numpy.array`, size (k,)
            The times at which to compute the solution
        inputs_dict : dict, optional
            Any input parameters to pass to the model when solving.
        """
        # Record whether there are any symbolic inputs
        inputs_dict = inputs_dict or {}

        # Create casadi objects for the root-finder
        inputs = casadi.vertcat(*[v for v in inputs_dict.values()])

        y0 = model.y0

        # The casadi algebraic solver can read rhs equations, but leaves them unchanged
        # i.e. the part of the solution vector that corresponds to the differential
        # equations will be equal to the initial condition provided. This allows this
        # solver to be used for initialising the DAE solvers
        if model.rhs == {}:
            len_rhs = 0
            y0_diff = casadi.DM()
            y0_alg = y0
        else:
            len_rhs = model.len_rhs
            y0_diff = y0[:len_rhs]
            y0_alg = y0[len_rhs:]

        y_alg = None

        # Set the parameter vector
        p = np.empty(1 + len_rhs + inputs.shape[0])
        # Set the time
        p[0] = 0
        # Set the differential states portion of the parameter vector
        p[1 : 1 + len_rhs] = np.asarray(y0_diff).ravel()
        # Set the inputs portion of the parameter vector
        p[1 + len_rhs :] = np.asarray(inputs).ravel()

        if getattr(model, "algebraic_root_solver", None) is None:
            self.set_up_root_solver(model, inputs_dict, t_eval)
        roots = model.algebraic_root_solver

        timer = pybamm.Timer()
        integration_time = 0
        for t in t_eval:
            # Set the time for the parameter vector
            p[0] = t

            # Solve
            success = False
            try:
                timer.reset()
                y_alg_sol = roots(y0_alg, p)
                integration_time += timer.time()

                # Check final output
                y_sol = casadi.vertcat(y0_diff, y_alg_sol)
                fun = model.casadi_algebraic(t, y_sol, inputs)
                # Casadi does not give us the value of the final residuals or step
                # norm, however, if it returns a success flag and there are no NaNs or Infs
                # in the solution, then it must be successful
                y_is_finite = np.isfinite(max_abs(y_alg_sol))
                fun_is_finite = np.isfinite(max_abs(fun))

                solver_stats = roots.stats()
                success = solver_stats["success"] and y_is_finite and fun_is_finite
            except RuntimeError as err:
                success = False
                message = err.args[0]
                raise pybamm.SolverError(
                    f"Could not find acceptable solution: {message}"
                ) from err

            if not success:
                if any(~np.isfinite(fun)):
                    reason = "solver returned NaNs or Infs"
                else:
                    reason = (
                        f"solver terminated unsuccessfully and maximum solution error ({casadi.mmax(casadi.fabs(fun))}) "
                        f"above tolerance ({self.tol})"
                    )

                raise pybamm.SolverError(
                    f"Could not find acceptable solution: {reason}"
                )

            # update initial guess for the next iteration
            y0_alg = y_alg_sol
            y0 = casadi.vertcat(y0_diff, y0_alg)
            # update solution array
            if y_alg is None:
                y_alg = y_alg_sol
            else:
                y_alg = casadi.horzcat(y_alg, y_alg_sol)

        # Concatenate differential part
        y_diff = casadi.horzcat(*[y0_diff] * len(t_eval))
        y_sol = casadi.vertcat(y_diff, y_alg)

        # Return solution object (no events, so pass None to t_event, y_event)
        sol = pybamm.Solution(
            [t_eval],
            y_sol,
            model,
            inputs_dict,
            termination="final time",
        )
        sol.integration_time = integration_time
        return sol


def max_abs(x):
    return np.linalg.norm(x, ord=np.inf)



#
# Summary Variable class
#
from __future__ import annotations
import pybamm
import numpy as np
from typing import Any


class SummaryVariables:
    """
    Class for managing and calculating summary variables from a PyBaMM solution.
    Summary variables are only calculated when simulations are run with PyBaMM
    Experiments.

    Parameters
    ----------
    solution : :class:`pybamm.Solution`
        The solution object to be used for creating the processed variables.
    cycle_summary_variables : list[pybamm.SummaryVariables], optional
        A list of cycle summary variables.
    esoh_solver : :class:`pybamm.lithium_ion.ElectrodeSOHSolver`, optional
        Solver for electrode state-of-health (eSOH) calculations.
    user_inputs : dict, optional
        Additional user inputs for calculations.

    Attributes
    ----------
    cycle_number : array[int]
        Stores the cycle number for each saved cycle, for use when plotting.
        Length is equal to the number of cycles in a solution.
    """

    def __init__(
        self,
        solution: pybamm.Solution,
        cycle_summary_variables: list[SummaryVariables] | None = None,
        esoh_solver: pybamm.lithium_ion.ElectrodeSOHSolver | None = None,
        user_inputs: dict[str, Any] | None = None,
    ):
        self.user_inputs = user_inputs or {}
        self.esoh_solver = esoh_solver
        self._variables = {}  # Store computed variables
        self.cycle_number = np.array([])

        model = solution.all_models[0]
        self._possible_variables = model.summary_variables  # minus esoh variables
        self._esoh_variables = None  # Store eSOH variable names

        # Flag if eSOH calculations are needed
        self.calc_esoh = (
            self.esoh_solver is not None
            and isinstance(model, pybamm.lithium_ion.BaseModel)
            and model.options.electrode_types["negative"] == "porous"
            and "Negative electrode capacity [A.h]" in model.variables
            and "Positive electrode capacity [A.h]" in model.variables
        )

        # Initialize based on cycle information
        if cycle_summary_variables:
            self._initialize_for_cycles(cycle_summary_variables)
        else:
            self.first_state = solution.first_state
            self.last_state = solution.last_state
            self.cycles = None

    def _initialize_for_cycles(self, cycle_summary_variables: list[SummaryVariables]):
        """Initialize attributes for when multiple cycles are provided."""
        self.first_state = None
        self.last_state = None
        self.cycles = cycle_summary_variables
        self.cycle_number = np.arange(1, len(self.cycles) + 1)
        first_cycle = self.cycles[0]
        self.calc_esoh = first_cycle.calc_esoh
        self.esoh_solver = first_cycle.esoh_solver
        self.user_inputs = first_cycle.user_inputs

    @property
    def all_variables(self) -> list[str]:
        """
        Return names of all possible summary variables, including eSOH variables
         if appropriate.
        """
        try:
            return self._all_variables
        except AttributeError:
            base_vars = self._possible_variables.copy()
            base_vars.extend(
                f"Change in {var[0].lower() + var[1:]}"
                for var in self._possible_variables
            )

            if self.calc_esoh:
                base_vars.extend(self.esoh_variables)

            self._all_variables = base_vars
            return self._all_variables

    @property
    def esoh_variables(self) -> list[str] | None:
        """Return names of all eSOH variables."""
        if self.calc_esoh and self._esoh_variables is None:
            esoh_model = self.esoh_solver._get_electrode_soh_sims_full().model
            esoh_vars = list(esoh_model.variables.keys())
            self._esoh_variables = esoh_vars
        return self._esoh_variables

    def __getitem__(self, key: str) -> float | list[float]:
        """
        Access or compute a summary variable by its name.

        Parameters
        ----------
        key : str
            The name of the variable

        Returns
        -------
        float or list[float]
        """

        if key in self._variables:
            # return it if it exists
            return self._variables[key]
        elif key == "Cycle number":
            return self.cycle_number
        elif key not in self.all_variables:
            # check it's listed as a summary variable
            raise KeyError(f"Variable '{key}' is not a summary variable.")
        else:
            # otherwise create it, save it and then return it
            if self.calc_esoh and key in self._esoh_variables:
                self.update_esoh()
            else:
                base_key = key.removeprefix("Change in ")
                base_key = base_key[0].upper() + base_key[1:]
                # this will create 'X' and 'Change in x' at the same time
                self.update(base_key)
            return self._variables[key]

    def update(self, var: str):
        """Compute and store a variable and its change."""
        var_lowercase = var[0].lower() + var[1:]
        if self.cycles:
            self._update_multiple_cycles(var, var_lowercase)
        else:
            self._update(var, var_lowercase)

    def _update_multiple_cycles(self, var: str, var_lowercase: str):
        """Creates aggregated summary variables for where more than one cycle exists."""
        var_cycle = [cycle[var] for cycle in self.cycles]
        change_var_cycle = [
            cycle[f"Change in {var_lowercase}"] for cycle in self.cycles
        ]
        self._variables[var] = var_cycle
        self._variables[f"Change in {var_lowercase}"] = change_var_cycle

    def _update(self, var: str, var_lowercase: str):
        """Create variable `var` for a single cycle."""
        data_first = self.first_state[var].data
        data_last = self.last_state[var].data
        self._variables[var] = data_last[0]
        self._variables[f"Change in {var_lowercase}"] = data_last[0] - data_first[0]

    def update_esoh(self):
        """Create all aggregated eSOH variables"""
        if self.cycles is not None:
            var_cycle = [cycle._get_esoh_variables() for cycle in self.cycles]
            aggregated_vars = {k: [] for k in var_cycle[0].keys()}
            for cycle in var_cycle:
                for k, v in cycle.items():
                    aggregated_vars[k].append(v)
            self._variables.update(aggregated_vars)
        else:
            self._variables.update(self._get_esoh_variables())

    def _get_esoh_variables(self) -> dict[str, float]:
        """Compute eSOH variables for a single solution."""
        Q_n = self.last_state["Negative electrode capacity [A.h]"].data[0]
        Q_p = self.last_state["Positive electrode capacity [A.h]"].data[0]
        Q_Li = self.last_state["Total lithium capacity in particles [A.h]"].data[0]
        all_inputs = {**self.user_inputs, "Q_n": Q_n, "Q_p": Q_p, "Q_Li": Q_Li}
        try:
            esoh_sol = self.esoh_solver.solve(inputs=all_inputs)
        except pybamm.SolverError as error:  # pragma: no cover
            raise pybamm.SolverError(
                "Could not solve for eSOH summary variables"
            ) from error

        return esoh_sol

    def get_summary_variables(self):
        """
        Computes and returns all the summary values + cycle number, as a dictionary
        """
        return {k: self[k] for k in [*self.all_variables, "Cycle number"]}



#
# Interface for discretisation
#
import pybamm
import numpy as np
from collections import defaultdict, OrderedDict
from scipy.sparse import block_diag, csc_matrix, csr_matrix
from scipy.sparse.linalg import inv


def has_bc_of_form(symbol, side, bcs, form):
    return (symbol in bcs) and (bcs[symbol][side][1] == form)


class Discretisation:
    """The discretisation class, with methods to process a model and replace
    Spatial Operators with Matrices and Variables with StateVectors

    Parameters
    ----------
    mesh : pybamm.Mesh
        contains all submeshes to be used on each domain
    spatial_methods : dict
        a dictionary of the spatial methods to be used on each
        domain. The keys correspond to the model domains and the
        values to the spatial method.
    check_model : bool, optional
            If True, model checks are performed after discretisation. For large
            systems these checks can be slow, so can be skipped by setting this
            option to False. When developing, testing or debugging it is recommended
            to leave this option as True as it may help to identify any errors.
            Default is True.
    remove_independent_variables_from_rhs : bool, optional
        If True, model checks to see whether any variables from the RHS are used
        in any other equation. If a variable meets all of the following criteria
        (not used anywhere in the model, len(rhs)>1), then the variable
        is moved to be explicitly integrated when called by the solution object.
        Default is False.
    """

    def __init__(
        self,
        mesh=None,
        spatial_methods=None,
        check_model=True,
        remove_independent_variables_from_rhs=False,
    ):
        self._mesh = mesh
        if mesh is None:
            self._spatial_methods = {}
        else:
            # Unpack macroscale to the constituent subdomains
            if "macroscale" in spatial_methods.keys():
                method = spatial_methods["macroscale"]
                spatial_methods["negative electrode"] = method
                spatial_methods["separator"] = method
                spatial_methods["positive electrode"] = method

            self._spatial_methods = spatial_methods
            for domain, method in self._spatial_methods.items():
                method.build(mesh)
                # Check zero-dimensional methods are only applied to zero-dimensional
                # meshes
                if isinstance(method, pybamm.ZeroDimensionalSpatialMethod):
                    if not isinstance(mesh[domain], pybamm.SubMesh0D):
                        raise pybamm.DiscretisationError(
                            "Zero-dimensional spatial method for the "
                            f"{domain} domain requires a zero-dimensional submesh"
                        )

        self._bcs = {}
        self.y_slices = {}
        self._discretised_symbols = {}
        self._check_model_flag = check_model
        self._remove_independent_variables_from_rhs_flag = (
            remove_independent_variables_from_rhs
        )

    @property
    def mesh(self):
        return self._mesh

    @property
    def y_slices(self):
        return self._y_slices

    @y_slices.setter
    def y_slices(self, value):
        if not isinstance(value, dict):
            raise TypeError(f"y_slices should be dict, not {type(value)}")

        self._y_slices = value

    @property
    def spatial_methods(self):
        return self._spatial_methods

    @property
    def bcs(self):
        return self._bcs

    @bcs.setter
    def bcs(self, value):
        self._bcs = value
        # reset discretised_symbols
        self._discretised_symbols = {}

    def process_model(self, model, inplace=True):
        """
        Discretise a model. Currently inplace, could be changed to return a new model.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            Model to dicretise. Must have attributes rhs, initial_conditions and
            boundary_conditions (all dicts of {variable: equation})
        inplace : bool, optional
            If True, discretise the model in place. Otherwise, return a new
            discretised model. Default is True.

        Returns
        -------
        model_disc : :class:`pybamm.BaseModel`
            The discretised model. Note that if ``inplace`` is True, model will
            have also been discretised in place so model == model_disc. If
            ``inplace`` is False, model != model_disc

        Raises
        ------
        :class:`pybamm.ModelError`
            If an empty model is passed (`model.rhs = {}` and `model.algebraic = {}` and
            `model.variables = {}`)

        """
        if model.is_discretised is True:
            raise pybamm.ModelError(
                "Cannot re-discretise a model. "
                "Set 'inplace=False' when first discretising a model to then be able "
                "to discretise it more times (e.g. for convergence studies)."
            )

        pybamm.logger.info(f"Start discretising {model.name}")

        # Make sure model isn't empty
        if (
            len(model.rhs) == 0
            and len(model.algebraic) == 0
            and len(model.variables) == 0
        ):
            raise pybamm.ModelError("Cannot discretise empty model")
        # Check well-posedness to avoid obscure errors
        model.check_well_posedness()

        # Prepare discretisation
        # set variables (we require the full variable not just id)

        # Search Equations for Independence
        if self._remove_independent_variables_from_rhs_flag:
            model = self.remove_independent_variables_from_rhs(model)
        variables = list(model.rhs.keys()) + list(model.algebraic.keys())
        # Find those RHS's that are constant
        if self.spatial_methods == {} and any(var.domain != [] for var in variables):
            for var in variables:
                if var.domain != []:
                    raise pybamm.DiscretisationError(
                        "Spatial method has not been given "
                        f"for variable {var.name} with domain {var.domain}"
                    )

        # Set the y split for variables
        pybamm.logger.verbose(f"Set variable slices for {model.name}")
        self.set_variable_slices(variables)

        # set boundary conditions (only need key ids for boundary_conditions)
        pybamm.logger.verbose(f"Discretise boundary conditions for {model.name}")
        self._bcs = self.process_boundary_conditions(model)
        pybamm.logger.verbose(f"Set internal boundary conditions for {model.name}")
        self.set_internal_boundary_conditions(model)

        # set up inplace vs not inplace
        if inplace:
            # any changes to model_disc attributes will change model attributes
            # since they point to the same object
            model_disc = model
        else:
            # create a copy of the original model
            model_disc = model.new_copy()

        # Keep a record of y_slices in the model
        model_disc.y_slices = self.y_slices_explicit
        # Keep a record of the bounds in the model
        model_disc.bounds = self.bounds

        model_disc.bcs = self.bcs

        # pre-process variables so that all state variables are included
        pre_processed_variables = self._pre_process_variables(
            model.variables, model.initial_conditions
        )

        pybamm.logger.verbose(f"Discretise initial conditions for {model.name}")
        ics, concat_ics = self.process_initial_conditions(model)
        model_disc.initial_conditions = ics
        model_disc.concatenated_initial_conditions = concat_ics

        # Discretise variables (applying boundary conditions)
        # Note that we **do not** discretise the keys of model.rhs,
        # model.initial_conditions and model.boundary_conditions
        pybamm.logger.verbose(f"Discretise variables for {model.name}")

        model_disc.variables = self.process_dict(pre_processed_variables)

        # Process parabolic and elliptic equations
        pybamm.logger.verbose(f"Discretise model equations for {model.name}")
        rhs, concat_rhs, alg, concat_alg = self.process_rhs_and_algebraic(model)
        model_disc.rhs, model_disc.concatenated_rhs = rhs, concat_rhs
        model_disc.algebraic, model_disc.concatenated_algebraic = alg, concat_alg

        # Save length of rhs and algebraic
        model_disc.len_rhs = model_disc.concatenated_rhs.size
        model_disc.len_alg = model_disc.concatenated_algebraic.size
        model_disc.len_rhs_and_alg = model_disc.len_rhs + model_disc.len_alg

        # Process events
        processed_events = []
        pybamm.logger.verbose(f"Discretise events for {model.name}")
        for event in model.events:
            pybamm.logger.debug(f"Discretise event '{event.name}'")
            processed_event = pybamm.Event(
                event.name, self.process_symbol(event.expression), event.event_type
            )
            processed_events.append(processed_event)
        model_disc.events = processed_events

        # Create mass matrix
        pybamm.logger.verbose(f"Create mass matrix for {model.name}")
        model_disc.mass_matrix, model_disc.mass_matrix_inv = self.create_mass_matrix(
            model_disc
        )

        # Save geometry
        pybamm.logger.verbose(f"Save geometry for {model.name}")
        model_disc._geometry = getattr(self.mesh, "_geometry", None)

        # Check that resulting model makes sense
        if self._check_model_flag:
            pybamm.logger.verbose(f"Performing model checks for {model.name}")
            self.check_model(model_disc)

        pybamm.logger.info(f"Finish discretising {model.name}")

        # Record that the model has been discretised
        model_disc.is_discretised = True

        return model_disc

    def set_variable_slices(self, variables):
        """
        Sets the slicing for variables.

        Parameters
        ----------
        variables : iterable of :class:`pybamm.Variables`
            The variables for which to set slices
        """
        # Set up y_slices and bounds
        y_slices = defaultdict(list)
        y_slices_explicit = defaultdict(list)
        start = 0
        end = 0
        lower_bounds = []
        upper_bounds = []
        # Iterate through unpacked variables, adding appropriate slices to y_slices
        for variable in variables:
            # Add up the size of all the domains in variable.domain
            if isinstance(variable, pybamm.ConcatenationVariable):
                start_ = start
                spatial_method = self.spatial_methods[variable.domain[0]]
                children = variable.children
                meshes = OrderedDict()
                for child in children:
                    meshes[child] = [spatial_method.mesh[dom] for dom in child.domain]
                sec_points = spatial_method._get_auxiliary_domain_repeats(
                    variable.domains
                )
                for _ in range(sec_points):
                    for child, mesh in meshes.items():
                        for domain_mesh in mesh:
                            end += domain_mesh.npts_for_broadcast_to_nodes
                        # Add to slices
                        y_slices[child].append(slice(start_, end))
                        y_slices_explicit[child].append(slice(start_, end))
                        # Increment start_
                        start_ = end
            else:
                end += self._get_variable_size(variable)

            # Add to slices
            y_slices[variable].append(slice(start, end))
            y_slices_explicit[variable].append(slice(start, end))

            # Add to bounds
            def evaluate_bound(bound, side):
                if bound.has_symbol_of_classes(pybamm.InputParameter):
                    if side == "lower":
                        return -np.inf
                    elif side == "upper":
                        return np.inf
                else:
                    return bound.evaluate()

            lower_bounds.extend(
                [evaluate_bound(variable.bounds[0], "lower")] * (end - start)
            )
            upper_bounds.extend(
                [evaluate_bound(variable.bounds[1], "upper")] * (end - start)
            )
            # Increment start
            start = end

        # Convert y_slices back to normal dictionary
        self.y_slices = dict(y_slices)
        # Also keep a record of what the y_slices are, to be stored in the model
        self.y_slices_explicit = dict(y_slices_explicit)

        # Also keep a record of bounds
        self.bounds = (np.array(lower_bounds), np.array(upper_bounds))

        # reset discretised_symbols
        self._discretised_symbols = {}

    def _get_variable_size(self, variable):
        """Helper function to determine what size a variable should be"""
        # If domain is empty then variable has size 1
        if variable.domain == []:
            return 1
        else:
            size = 0
            spatial_method = self.spatial_methods[variable.domain[0]]
            repeats = spatial_method._get_auxiliary_domain_repeats(variable.domains)
            for dom in variable.domain:
                size += spatial_method.mesh[dom].npts_for_broadcast_to_nodes * repeats
            return size

    def set_internal_boundary_conditions(self, model):
        """
        A method to set the internal boundary conditions for the submodel.
        These are required to properly calculate the gradient.
        Note: this method modifies the state of self.boundary_conditions.
        """

        def boundary_gradient(left_symbol, right_symbol):
            pybamm.logger.debug(
                f"Calculate boundary gradient ({left_symbol} and {right_symbol})"
            )
            left_domain = left_symbol.domain[0]
            right_domain = right_symbol.domain[0]

            left_mesh = self.spatial_methods[left_domain].mesh[left_domain]
            right_mesh = self.spatial_methods[right_domain].mesh[right_domain]

            left_symbol_disc = self.process_symbol(left_symbol)
            right_symbol_disc = self.process_symbol(right_symbol)

            return self.spatial_methods[left_domain].internal_neumann_condition(
                left_symbol_disc, right_symbol_disc, left_mesh, right_mesh
            )

        bc_keys = list(self.bcs.keys())

        internal_bcs = {}
        for var in model.boundary_conditions.keys():
            if isinstance(var, pybamm.Concatenation):
                children = var.orphans

                first_child = children[0]
                next_child = children[1]

                lbc = self.bcs[var]["left"]
                rbc = (boundary_gradient(first_child, next_child), "Neumann")

                if first_child not in bc_keys:
                    internal_bcs.update({first_child: {"left": lbc, "right": rbc}})

                for current_child, next_child in zip(children[1:-1], children[2:]):
                    lbc = rbc
                    rbc = (boundary_gradient(current_child, next_child), "Neumann")
                    if current_child not in bc_keys:
                        internal_bcs.update(
                            {current_child: {"left": lbc, "right": rbc}}
                        )

                lbc = rbc
                rbc = self.bcs[var]["right"]
                if children[-1] not in bc_keys:
                    internal_bcs.update({children[-1]: {"left": lbc, "right": rbc}})

        self.bcs.update(internal_bcs)

    def process_initial_conditions(self, model):
        """Discretise model initial_conditions.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            Model to dicretise. Must have attributes rhs, initial_conditions and
            boundary_conditions (all dicts of {variable: equation})

        Returns
        -------
        tuple
            Tuple of processed_initial_conditions (dict of initial conditions) and
            concatenated_initial_conditions (numpy array of concatenated initial
            conditions)

        """
        # Discretise initial conditions
        processed_initial_conditions = self.process_dict(
            model.initial_conditions, ics=True
        )

        # Concatenate initial conditions into a single vector
        # check that all initial conditions are set
        processed_concatenated_initial_conditions = self._concatenate_in_order(
            processed_initial_conditions, check_complete=True
        )

        return processed_initial_conditions, processed_concatenated_initial_conditions

    def process_boundary_conditions(self, model):
        """Discretise model boundary_conditions, also converting keys to ids

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            Model to dicretise. Must have attributes rhs, initial_conditions and
            boundary_conditions (all dicts of {variable: equation})

        Returns
        -------
        dict
            Dictionary of processed boundary conditions

        """

        processed_bcs = {}

        # process and set pybamm.variables first incase required
        # in discrisation of other boundary conditions
        for key, bcs in model.boundary_conditions.items():
            processed_bcs[key] = {}

            # check if the boundary condition at the origin for sphere domains is other
            # than no flux
            for subdomain in key.domain:
                if (
                    self.mesh[subdomain].coord_sys
                    in ["spherical polar", "cylindrical polar"]
                    and next(iter(self.mesh.geometry[subdomain].values()))["min"] == 0
                ):
                    if bcs["left"][0].value != 0 or bcs["left"][1] != "Neumann":
                        raise pybamm.ModelError(
                            "Boundary condition at r = 0 must be a homogeneous "
                            f"Neumann condition for {self.mesh[subdomain].coord_sys} coordinates"
                        )

            # Handle any boundary conditions applied on the tabs
            if any("tab" in side for side in list(bcs.keys())):
                bcs = self.check_tab_conditions(key, bcs)

            # Process boundary conditions
            for side, bc in bcs.items():
                eqn, typ = bc
                pybamm.logger.debug(f"Discretise {key} ({side} bc)")
                processed_eqn = self.process_symbol(eqn)
                processed_bcs[key][side] = (processed_eqn, typ)

        return processed_bcs

    def check_tab_conditions(self, symbol, bcs):
        """
        Check any boundary conditions applied on "negative tab", "positive tab"
        and "no tab". For 1D current collector meshes, these conditions are
        converted into boundary conditions on "left" (tab at z=0) or "right"
        (tab at z=l_z) depending on the tab location stored in the mesh. For 2D
        current collector meshes, the boundary conditions can be applied on the
        tabs directly.

        Parameters
        ----------
        symbol : :class:`pybamm.expression_tree.symbol.Symbol`
            The symbol on which the boundary conditions are applied.
        bcs : dict
            The dictionary of boundary conditions (a dict of {side: equation}).

        Returns
        -------
        dict
            The dictionary of boundary conditions, with the keys changed to
            "left" and "right" where necessary.

        """
        # Check symbol domain
        domain = symbol.domain[0]
        mesh = self.mesh[domain]

        if domain != "current collector":
            raise pybamm.ModelError(
                "Boundary conditions can only be applied on the tabs in the domain "
                f"'current collector', but {symbol} has domain {domain}"
            )
        # Replace keys with "left" and "right" as appropriate for 1D meshes
        if isinstance(mesh, pybamm.SubMesh1D):
            # send boundary conditions applied on the tabs to "left" or "right"
            # depending on the tab location stored in the mesh
            for tab in ["negative tab", "positive tab"]:
                if any(tab in side for side in list(bcs.keys())):
                    bcs[mesh.tabs[tab]] = bcs.pop(tab)
            # if there was a tab at either end, then the boundary conditions
            # have now been set on "left" and "right" as required by the spatial
            # method, so there is no need to further modify the bcs dict
            if all(side in list(bcs.keys()) for side in ["left", "right"]):
                pass
            # if both tabs are located at z=0 then the "right" boundary condition
            # (at z=1) is the condition for "no tab"
            elif "left" in list(bcs.keys()):
                bcs["right"] = bcs.pop("no tab")
            # else if both tabs are located at z=1, the "left" boundary condition
            # (at z=0) is the condition for "no tab"
            else:
                bcs["left"] = bcs.pop("no tab")

        return bcs

    def process_rhs_and_algebraic(self, model):
        """Discretise model equations - differential ('rhs') and algebraic.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            Model to dicretise. Must have attributes rhs, initial_conditions and
            boundary_conditions (all dicts of {variable: equation})

        Returns
        -------
        tuple
            Tuple of processed_rhs (dict of processed differential equations),
            processed_concatenated_rhs, processed_algebraic (dict of processed algebraic
            equations) and processed_concatenated_algebraic

        """

        # Discretise right-hand sides, passing domain from variable
        processed_rhs = self.process_dict(model.rhs)

        # Concatenate rhs into a single state vector
        # Need to concatenate in order as the ordering of equations could be different
        # in processed_rhs and model.rhs
        processed_concatenated_rhs = self._concatenate_in_order(processed_rhs)

        # Discretise and concatenate algebraic equations
        processed_algebraic = self.process_dict(model.algebraic)

        # Concatenate algebraic into a single state vector
        # Need to concatenate in order as the ordering of equations could be different
        # in processed_algebraic and model.algebraic
        processed_concatenated_algebraic = self._concatenate_in_order(
            processed_algebraic
        )

        return (
            processed_rhs,
            processed_concatenated_rhs,
            processed_algebraic,
            processed_concatenated_algebraic,
        )

    def create_mass_matrix(self, model):
        """Creates mass matrix of the discretised model.
        Note that the model is assumed to be of the form M*y_dot = f(t,y), where
        M is the (possibly singular) mass matrix.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            Discretised model. Must have attributes rhs, initial_conditions and
            boundary_conditions (all dicts of {variable: equation})

        Returns
        -------
        :class:`pybamm.Matrix`
            The mass matrix
        :class:`pybamm.Matrix`
            The inverse of the ode part of the mass matrix (required by solvers
            which only accept the ODEs in explicit form)
        """
        # Create list of mass matrices for each equation to be put into block
        # diagonal mass matrix for the model
        mass_list = []
        mass_inv_list = []

        # get a list of model rhs variables that are sorted according to
        # where they are in the state vector
        model_variables = model.rhs.keys()
        model_slices = []
        for v in model_variables:
            model_slices.append(self.y_slices[v][0])
        sorted_model_variables = [
            v for _, v in sorted(zip(model_slices, model_variables))
        ]

        # Process mass matrices for the differential equations
        for var in sorted_model_variables:
            if var.domain == []:
                # If variable domain empty then mass matrix is just 1
                mass = 1.0
                mass_inv = 1.0
            else:
                mass = (
                    self.spatial_methods[var.domain[0]]
                    .mass_matrix(var, self.bcs)
                    .entries
                )
                if isinstance(
                    self.spatial_methods[var.domain[0]],
                    (pybamm.ZeroDimensionalSpatialMethod, pybamm.FiniteVolume),
                ):
                    # for 0D methods the mass matrix is just a scalar 1 and for
                    # finite volumes the mass matrix is identity, so no need to
                    # compute the inverse
                    mass_inv = mass
                else:
                    # inverse is more efficient in csc format
                    mass_inv = inv(csc_matrix(mass))

            mass_list.append(mass)
            mass_inv_list.append(mass_inv)

        # Create lumped mass matrix (of zeros) of the correct shape for the
        # discretised algebraic equations
        if model.algebraic.keys():
            mass_algebraic_size = model.concatenated_algebraic.shape[0]
            mass_algebraic = csr_matrix((mass_algebraic_size, mass_algebraic_size))
            mass_list.append(mass_algebraic)

        # Create block diagonal (sparse) mass matrix (if model is not empty)
        # and inverse (if model has odes)
        N_rhs = len(model.rhs)
        N_alg = len(model.algebraic)

        has_mass_matrix = N_rhs > 0 or N_alg > 0
        has_mass_matrix_inv = N_rhs > 0

        if has_mass_matrix:
            mass_matrix = pybamm.Matrix(block_diag(mass_list, format="csr"))
        else:
            mass_matrix = None

        if has_mass_matrix_inv:
            mass_matrix_inv = pybamm.Matrix(block_diag(mass_inv_list, format="csr"))
        else:
            mass_matrix_inv = None

        return mass_matrix, mass_matrix_inv

    def _pre_process_variables(
        self,
        variables: dict[str, pybamm.Symbol],
        initial_conditions: dict[pybamm.Variable, pybamm.Symbol],
    ):
        """
        Pre-process variables before discretisation. This involves:
        - ensuring that all the state variables are included in the variables,
          any missing are added

        Parameters
        ----------
        variables : dict
            Dictionary of variables to pre-process
        initial_conditions : dict
            Dictionary of initial conditions

        Returns
        -------
        dict
            Pre-processed variables (copy of input variables with any missing state)

        Raises
        ------
        :class:`pybamm.ModelError`
            If any state variable names are already included but with
            incorrect expressions
        """
        new_variables = {k: v for k, v in variables.items()}
        for var in initial_conditions.keys():
            if var.name not in new_variables:
                new_variables[var.name] = var
            else:
                if new_variables[var.name] != var:
                    raise pybamm.ModelError(
                        f"Variable '{var.name}' should have expression "
                        f"'{var}', but has expression '{new_variables[var.name]}'"
                    )
        return new_variables

    def process_dict(self, var_eqn_dict, ics=False):
        """Discretise a dictionary of {variable: equation}, broadcasting if necessary
        (can be model.rhs, model.algebraic, model.initial_conditions or
        model.variables).

        Parameters
        ----------
        var_eqn_dict : dict
            Equations ({variable: equation} dict) to dicretise
            (can be model.rhs, model.algebraic, model.initial_conditions or
            model.variables)
        ics : bool, optional
            Whether the equations are initial conditions. If True, the equations are
            scaled by the reference value of the variable, if given

        Returns
        -------
        new_var_eqn_dict : dict
            Discretised equations

        """
        new_var_eqn_dict = {}
        for eqn_key, eqn in var_eqn_dict.items():
            # Broadcast if the equation evaluates to a number (e.g. Scalar)
            if np.prod(eqn.shape_for_testing) == 1 and not isinstance(eqn_key, str):
                if eqn_key.domain == []:
                    eqn = eqn * pybamm.Vector([1])
                else:
                    eqn = pybamm.FullBroadcast(eqn, broadcast_domains=eqn_key.domains)

            pybamm.logger.debug(f"Discretise {eqn_key!r}")
            processed_eqn = self.process_symbol(eqn)
            # Calculate scale if the key has a scale
            scale = getattr(eqn_key, "scale", 1)
            if ics:
                reference = getattr(eqn_key, "reference", 0)
            else:
                reference = 0

            if scale != 1 or reference != 0:
                processed_eqn = (processed_eqn - reference) / scale

            new_var_eqn_dict[eqn_key] = processed_eqn
        return new_var_eqn_dict

    def process_symbol(self, symbol):
        """Discretise operators in model equations.
        If a symbol has already been discretised, the stored value is returned.

        Parameters
        ----------
        symbol : :class:`pybamm.expression_tree.symbol.Symbol`
            Symbol to discretise

        Returns
        -------
        :class:`pybamm.expression_tree.symbol.Symbol`
            Discretised symbol

        """
        try:
            return self._discretised_symbols[symbol]
        except KeyError:
            discretised_symbol = self._process_symbol(symbol)
            self._discretised_symbols[symbol] = discretised_symbol
            discretised_symbol.test_shape()

            # Assign mesh as an attribute to the processed variable
            if symbol.domain != []:
                discretised_symbol.mesh = self.mesh[symbol.domain]
            else:
                discretised_symbol.mesh = None

            # Assign secondary mesh
            if symbol.domains["secondary"] != []:
                discretised_symbol.secondary_mesh = self.mesh[
                    symbol.domains["secondary"]
                ]
            else:
                discretised_symbol.secondary_mesh = None

            # Assign tertiary mesh
            if symbol.domains["tertiary"] != []:
                discretised_symbol.tertiary_mesh = self.mesh[symbol.domains["tertiary"]]
            else:
                discretised_symbol.tertiary_mesh = None

            return discretised_symbol

    def _process_symbol(self, symbol):
        """See :meth:`Discretisation.process_symbol()`."""

        if symbol.domain != []:
            spatial_method = self.spatial_methods[symbol.domain[0]]
            # If boundary conditions are provided, need to check for BCs on tabs
            if self.bcs:
                key_id = next(iter(self.bcs.keys()))
                if any("tab" in side for side in list(self.bcs[key_id].keys())):
                    self.bcs[key_id] = self.check_tab_conditions(
                        symbol, self.bcs[key_id]
                    )

        if isinstance(symbol, pybamm.BinaryOperator):
            # Pre-process children
            left, right = symbol.children
            disc_left = self.process_symbol(left)
            disc_right = self.process_symbol(right)
            if symbol.domain == []:
                return pybamm.simplify_if_constant(
                    symbol.create_copy(new_children=[disc_left, disc_right])
                )
            else:
                return spatial_method.process_binary_operators(
                    symbol, left, right, disc_left, disc_right
                )
        elif isinstance(symbol, pybamm._BaseAverage):
            # Create a new Integral operator and process it
            child = symbol.orphans[0]
            if isinstance(symbol, pybamm.SizeAverage):
                R = symbol.integration_variable[0]
                f_a_dist = symbol.f_a_dist
                # take average using Integral and distribution f_a_dist
                average = pybamm.Integral(f_a_dist * child, R) / pybamm.Integral(
                    f_a_dist, R
                )
            else:
                x = symbol.integration_variable
                v = pybamm.ones_like(child)
                average = pybamm.Integral(child, x) / pybamm.Integral(v, x)
            return self.process_symbol(average)

        elif isinstance(symbol, pybamm.UnaryOperator):
            child = symbol.child

            disc_child = self.process_symbol(child)
            if child.domain != []:
                child_spatial_method = self.spatial_methods[child.domain[0]]

            if isinstance(symbol, pybamm.Gradient):
                return child_spatial_method.gradient(child, disc_child, self.bcs)

            elif isinstance(symbol, pybamm.Divergence):
                return child_spatial_method.divergence(child, disc_child, self.bcs)

            elif isinstance(symbol, pybamm.Laplacian):
                return child_spatial_method.laplacian(child, disc_child, self.bcs)

            elif isinstance(symbol, pybamm.GradientSquared):
                return child_spatial_method.gradient_squared(
                    child, disc_child, self.bcs
                )

            elif isinstance(symbol, pybamm.Mass):
                return child_spatial_method.mass_matrix(child, self.bcs)

            elif isinstance(symbol, pybamm.BoundaryMass):
                return child_spatial_method.boundary_mass_matrix(child, self.bcs)

            elif isinstance(symbol, pybamm.IndefiniteIntegral):
                return child_spatial_method.indefinite_integral(
                    child, disc_child, "forward"
                )
            elif isinstance(symbol, pybamm.BackwardIndefiniteIntegral):
                return child_spatial_method.indefinite_integral(
                    child, disc_child, "backward"
                )

            elif isinstance(symbol, pybamm.Integral):
                integral_spatial_method = self.spatial_methods[
                    symbol.integration_variable[0].domain[0]
                ]
                out = integral_spatial_method.integral(
                    child, disc_child, symbol._integration_dimension
                )
                out.copy_domains(symbol)
                return out

            elif isinstance(symbol, pybamm.DefiniteIntegralVector):
                return child_spatial_method.definite_integral_matrix(
                    child, vector_type=symbol.vector_type
                )

            elif isinstance(symbol, pybamm.BoundaryIntegral):
                return child_spatial_method.boundary_integral(
                    child, disc_child, symbol.region
                )

            elif isinstance(symbol, pybamm.Broadcast):
                # Broadcast new_child to the domain specified by symbol.domain
                # Different discretisations may broadcast differently
                return spatial_method.broadcast(
                    disc_child, symbol.domains, symbol.broadcast_type
                )

            elif isinstance(symbol, pybamm.DeltaFunction):
                return spatial_method.delta_function(symbol, disc_child)

            elif isinstance(symbol, pybamm.BoundaryOperator):
                # if boundary operator applied on "negative tab" or
                # "positive tab" *and* the mesh is 1D then change side to
                # "left" or "right" as appropriate
                if symbol.side in ["negative tab", "positive tab"]:
                    mesh = self.mesh[symbol.children[0].domain[0]]
                    if isinstance(mesh, pybamm.SubMesh1D):
                        symbol.side = mesh.tabs[symbol.side]
                return child_spatial_method.boundary_value_or_flux(
                    symbol, disc_child, self.bcs
                )
            elif isinstance(symbol, pybamm.EvaluateAt):
                return child_spatial_method.evaluate_at(
                    symbol, disc_child, symbol.position
                )
            elif isinstance(symbol, pybamm.UpwindDownwind):
                direction = symbol.name  # upwind or downwind
                return spatial_method.upwind_or_downwind(
                    child, disc_child, self.bcs, direction
                )
            elif isinstance(symbol, pybamm.NotConstant):
                # After discretisation, we can make the symbol constant
                return disc_child
            else:
                return symbol.create_copy(new_children=[disc_child])

        elif isinstance(symbol, pybamm.Function):
            disc_children = [self.process_symbol(child) for child in symbol.children]
            return symbol.create_copy(disc_children)

        elif isinstance(symbol, pybamm.VariableDot):
            # Add symbol's reference and multiply by the symbol's scale
            # so that the state vector is of order 1
            return symbol.reference + symbol.scale * pybamm.StateVectorDot(
                *self.y_slices[symbol.get_variable()],
                domains=symbol.domains,
            )

        elif isinstance(symbol, pybamm.Variable):
            # add a try except block for a more informative error if a variable
            # can't be found. This should usually be caught earlier by
            # model.check_well_posedness, but won't be if debug_mode is False
            try:
                y_slices = self.y_slices[symbol]
            except KeyError as error:
                raise pybamm.ModelError(
                    f"No key set for variable '{symbol.name}'. Make sure it is included in either "
                    "model.rhs or model.algebraic in an unmodified form "
                    "(e.g. not Broadcasted)"
                ) from error
            # Add symbol's reference and multiply by the symbol's scale
            # so that the state vector is of order 1
            return symbol.reference + symbol.scale * pybamm.StateVector(
                *y_slices, domains=symbol.domains
            )

        elif isinstance(symbol, pybamm.SpatialVariable):
            return spatial_method.spatial_variable(symbol)

        elif isinstance(symbol, pybamm.ConcatenationVariable):
            # create new children without scale and reference
            # the scale and reference will be applied to the concatenation instead
            new_children = []
            old_y_slices = self.y_slices.copy()
            for child in symbol.children:
                child_no_scale = child.create_copy()
                child_no_scale._scale = 1
                child_no_scale._reference = 0
                child_no_scale.set_id()
                self.y_slices[child_no_scale] = self.y_slices[child]
                new_children.append(self.process_symbol(child_no_scale))
            self.y_slices = old_y_slices
            new_symbol = spatial_method.concatenation(new_children)
            # apply scale to the whole concatenation
            return symbol.reference + symbol.scale * new_symbol

        elif isinstance(symbol, pybamm.Concatenation):
            new_children = [self.process_symbol(child) for child in symbol.children]
            new_symbol = spatial_method.concatenation(new_children)
            return new_symbol

        elif isinstance(symbol, pybamm.InputParameter):
            if symbol.domain != []:
                expected_size = self._get_variable_size(symbol)
            else:
                expected_size = None
            if symbol._expected_size is None:
                symbol._expected_size = expected_size
            return symbol.create_copy()

        elif isinstance(symbol, pybamm.CoupledVariable):
            new_symbol = self.process_symbol(symbol.children[0])
            return new_symbol

        else:
            # Backup option: return the object
            return symbol

    def concatenate(self, *symbols, sparse=False):
        if sparse:
            return pybamm.SparseStack(*symbols)
        else:
            return pybamm.numpy_concatenation(*symbols)

    def _concatenate_in_order(self, var_eqn_dict, check_complete=False, sparse=False):
        """
        Concatenate a dictionary of {variable: equation} using self.y_slices

        The keys/variables in `var_eqn_dict` must be the same as the ids in
        `self.y_slices`.
        The resultant concatenation is ordered according to the ordering of the slice
        values in `self.y_slices`

        Parameters
        ----------
        var_eqn_dict : dict
            Equations ({variable: equation} dict) to dicretise
        check_complete : bool, optional
            Whether to check keys in var_eqn_dict against self.y_slices. Default
            is False
        sparse : bool, optional
            If True the concatenation will be a :class:`pybamm.SparseStack`. If
            False the concatenation will be a :class:`pybamm.NumpyConcatenation`.
            Default is False

        Returns
        -------
        var_eqn_dict : dict
            Discretised right-hand side equations

        """
        # Unpack symbols in variables that are concatenations of variables
        unpacked_variables = []
        slices = []
        for symbol in var_eqn_dict.keys():
            if isinstance(symbol, pybamm.ConcatenationVariable):
                unpacked_variables.extend([symbol] + [var for var in symbol.children])
            else:
                unpacked_variables.append(symbol)
            slices.append(self.y_slices[symbol][0])

        if check_complete:
            # Check keys from the given var_eqn_dict against self.y_slices
            unpacked_variables_set = set(unpacked_variables)
            if unpacked_variables_set != set(self.y_slices.keys()):
                given_variable_names = [v.name for v in var_eqn_dict.keys()]
                raise pybamm.ModelError(
                    "Initial conditions are insufficient. Only "
                    f"provided for {given_variable_names} "
                )

        equations = list(var_eqn_dict.values())

        # sort equations according to slices
        sorted_equations = [eq for _, eq in sorted(zip(slices, equations))]

        return self.concatenate(*sorted_equations, sparse=sparse)

    def check_model(self, model):
        """Perform some basic checks to make sure the discretised model makes sense."""
        self.check_initial_conditions(model)

    def check_initial_conditions(self, model):
        # Check initial conditions are a numpy array
        # Individual
        for var, eqn in model.initial_conditions.items():
            ic_eval = eqn.evaluate(t=0, inputs="shape test")
            if not isinstance(ic_eval, np.ndarray):
                raise pybamm.ModelError(
                    "initial conditions must be numpy array after discretisation but "
                    f"they are {type(ic_eval)} for variable '{var}'."
                )

            # Check that the initial condition is within the bounds
            # Skip this check if there are input parameters in the initial conditions
            bounds = var.bounds
            if not eqn.has_symbol_of_classes(pybamm.InputParameter) and not (
                all(bounds[0].value <= ic_eval) and all(ic_eval <= bounds[1].value)
            ):
                raise pybamm.ModelError(
                    "initial condition is outside of variable bounds "
                    f"{bounds} for variable '{var}'."
                )

        # Check initial conditions and model equations have the same shape
        # Individual
        for var in model.rhs.keys():
            if model.rhs[var].shape != model.initial_conditions[var].shape:
                raise pybamm.ModelError(
                    "rhs and initial conditions must have the same shape after "
                    "discretisation but rhs.shape = "
                    f"{model.rhs[var].shape} and initial_conditions.shape = {model.initial_conditions[var].shape} for variable '{var}'."
                )
        for var in model.algebraic.keys():
            if model.algebraic[var].shape != model.initial_conditions[var].shape:
                raise pybamm.ModelError(
                    "algebraic and initial conditions must have the same shape after "
                    "discretisation but algebraic.shape = "
                    f"{model.algebraic[var].shape} and initial_conditions.shape = {model.initial_conditions[var].shape} for variable '{var}'."
                )

    def is_variable_independent(self, var, all_vars_in_eqns):
        pybamm.logger.verbose("Removing independent blocks.")
        if not isinstance(var, pybamm.Variable):
            return False

        this_var_is_independent = var.name not in all_vars_in_eqns
        not_in_y_slices = var not in list(self.y_slices.keys())
        not_in_discretised = var not in list(self._discretised_symbols.keys())
        is_0D = len(var.domain) == 0
        this_var_is_independent = (
            this_var_is_independent and not_in_y_slices and not_in_discretised and is_0D
        )
        return this_var_is_independent

    def remove_independent_variables_from_rhs(self, model):
        rhs_vars_to_search_over = list(model.rhs.keys())
        unpacker = pybamm.SymbolUnpacker(pybamm.Variable)
        eqns_to_check = (
            list(model.rhs.values())
            + list(model.algebraic.values())
            + [
                x[side][0]
                for x in model.boundary_conditions.values()
                for side in x.keys()
            ]
            # only check children of variables, this will skip the variable itself
            # and catch any other cases
            + [child for var in model.variables.values() for child in var.children]
            + [event.expression for event in model.events]
        )
        all_vars_in_eqns = unpacker.unpack_list_of_symbols(eqns_to_check)
        all_vars_in_eqns = [var.name for var in all_vars_in_eqns]

        for var in rhs_vars_to_search_over:
            this_var_is_independent = self.is_variable_independent(
                var, all_vars_in_eqns
            )
            if this_var_is_independent:
                if len(model.rhs) != 1:
                    pybamm.logger.info(f"removing variable {var} from rhs")
                    my_initial_condition = model.initial_conditions[var]
                    model.variables[var.name] = pybamm.ExplicitTimeIntegral(
                        model.rhs[var], my_initial_condition
                    )
                    # edge case where a variable appears
                    # in variables twice under different names
                    for key in model.variables:
                        if model.variables[key] == var:
                            print("here")
                            model.variables[key] = model.variables[var.name]
                    del model.rhs[var]
                    del model.initial_conditions[var]
                else:
                    break
        return model
